# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Instrument Controllers class"""

import io
from dataclasses import dataclass

from ruamel.yaml import YAML

from qililab.instrument_controllers.instrument_controller import InstrumentController


@dataclass
class InstrumentControllers:
    """Instrument Controllers class."""

    elements: list[InstrumentController]

    def get_instrument_controller(self, alias: str):
        """Get instrument controller given an id and category"""
        return next((instrument for instrument in self.elements if instrument.alias == alias), None)

    def connect(self):
        """Connect to all instrument controllers."""
        for instrument_controller in self.elements:
            instrument_controller.connect()

    def initial_setup(self):
        """Set the initial setup of the instruments"""
        for instrument_controller in self.elements:
            instrument_controller.initial_setup()

    def turn_on_instruments(self):
        """Turn on the instrument"""
        for instrument_controller in self.elements:
            instrument_controller.turn_on()

    def turn_off_instruments(self):
        """Turn off the instrument"""
        for instrument_controller in self.elements:
            instrument_controller.turn_off()

    def disconnect(self):
        """Disconnect from all instrument controllers."""
        for instrument_controller in self.elements:
            instrument_controller.disconnect()

    def to_dict(self):
        """Return a dict representation of the Instrument Controllers class."""
        return [instrument_controller.to_dict() for instrument_controller in self.elements]

    def __str__(self) -> str:
        """
        Returns:
            str: String representation of the Instrument Controllers class.
        """
        return str(YAML().dump(self.to_dict(), io.BytesIO()))



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Instrument Controllers module."""

from .instrument_controller import InstrumentController
from .instrument_controllers import InstrumentControllers
from .keithley import Keithley2600Controller
from .mini_circuits import MiniCircuitsController
from .qblox import QbloxClusterController, QbloxSPIRackController
from .qdevil import QDevilQDac2Controller
from .quantum_machines import QuantumMachinesClusterController
from .rohde_schwarz import SGS100AController
from .single_instrument_controller import SingleInstrumentController
from .utils import InstrumentControllerFactory
from .yokogawa import GS200Controller

__all__ = [
    "GS200Controller",
    "InstrumentController",
    "InstrumentControllerFactory",
    "InstrumentControllers",
    "Keithley2600Controller",
    "MiniCircuitsController",
    "QDevilQDac2Controller",
    "QbloxClusterController",
    "QbloxSPIRackController",
    "QuantumMachinesClusterController",
    "SGS100AController",
    "SingleInstrumentController",
]



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Keithley2600 instrument."""

from dataclasses import dataclass

import numpy as np

from qililab.instruments.instrument import Instrument, ParameterNotFound
from qililab.instruments.utils import InstrumentFactory
from qililab.typings import ChannelID, InstrumentName, Keithley2600Driver, Parameter, ParameterValue


@InstrumentFactory.register
class Keithley2600(Instrument):
    """Keithley2600 class.
    Args:
        name (InstrumentName): name of the instrument
        device (Keithley2600Driver): Instance of the Instrument device driver.
        settings (Keithley2600Settings): Settings of the instrument.
    """

    name = InstrumentName.KEITHLEY2600

    @dataclass
    class Keithley2600Settings(Instrument.InstrumentSettings):
        """Settings for Keithley2600 instrument."""

        max_current: float
        max_voltage: float

    settings: Keithley2600Settings
    device: Keithley2600Driver

    def set_parameter(self, parameter: Parameter, value: ParameterValue, channel_id: ChannelID | None = None):
        """Setup instrument."""
        if parameter == Parameter.MAX_CURRENT:
            self.max_current = float(value)
            if self.is_device_active():
                self.device.smua.limiti(self.max_current)
            return
        if parameter == Parameter.MAX_VOLTAGE:
            self.max_voltage = float(value)
            if self.is_device_active():
                self.device.smua.limitv(self.max_voltage)
            return
        raise ParameterNotFound(self, parameter)

    def get_parameter(self, parameter: Parameter, channel_id: ChannelID | None = None):
        """Setup instrument."""
        if parameter == Parameter.MAX_CURRENT:
            return self.max_current
        if parameter == Parameter.MAX_VOLTAGE:
            return self.max_voltage
        raise ParameterNotFound(self, parameter)

    def initial_setup(self):
        """performs an initial setup"""
        self.device.smua.limiti(self.max_current)
        self.device.smua.limitv(self.max_voltage)

    def turn_on(self):
        """Turn on an instrument."""

    def turn_off(self):
        """Turn off an instrument."""

    def reset(self):
        """Reset instrument."""
        self.device.reset()

    def fast_sweep(self, start: float, stop: float, steps: int, mode: str) -> tuple[np.ndarray, np.ndarray]:
        """Perform a fast sweep using a deployed lua script and return an xarray dataset.

        Args:
            start: starting sweep value (V or A)
            stop: end sweep value (V or A)
            steps: number of steps
            mode: Type of sweep, either 'IV' (voltage sweep), 'VI' (current sweep two probe setup) or 'VIfourprobe'
            (current sweep four probe setup)

        Returns:
            np.ndarray: Measured data.
        """
        data = self.device.smua.doFastSweep(start=start, stop=stop, steps=steps, mode=mode)
        x_values = np.linspace(start=start, stop=stop, num=steps)
        return x_values, data.to_xarray().to_array().values.squeeze()

    @property
    def max_current(self):
        """Keithley2600 'max_current' property.

        Returns:
            float: Maximum current allowed in voltage mode.
        """
        return self.settings.max_current

    @max_current.setter
    def max_current(self, value: float):
        """Keithley2600 'max_current' setter property.

        Args:
            float: Maximum current allowed in voltage mode.
        """
        if self.is_device_active():
            self.device.smua.limiti(value)
        self.settings.max_current = value

    @property
    def max_voltage(self):
        """Keithley2600 'max_voltage' property.

        Returns:
            float: Maximum voltage allowed in current mode.
        """
        return self.settings.max_voltage

    @max_voltage.setter
    def max_voltage(self, value: float):
        """Keithley2600 'max_voltage' setter property.

        Args:
            float: Maximum voltage allowed in current mode.
        """
        if self.is_device_active():
            self.device.smua.limitv(value)
        self.settings.max_voltage = value



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import base64
import types
from collections import deque
from uuid import UUID

import numpy as np
from dill import dumps, loads  # noqa: S403
from ruamel.yaml import YAML


def ndarray_representer(representer, data):
    """Representer for ndarray"""
    value = {"dtype": str(data.dtype), "shape": data.shape, "data": data.ravel().tolist()}
    return representer.represent_mapping("!ndarray", value)


def ndarray_constructor(constructor, node):
    """Constructor for ndarray"""
    mapping = constructor.construct_mapping(node, deep=True)
    dtype = np.dtype(mapping["dtype"])
    shape = tuple(mapping["shape"])
    data = mapping["data"]
    return np.array(data, dtype=dtype).reshape(shape)


def deque_representer(representer, data):
    """Representer for deque"""
    return representer.represent_sequence("!deque", list(data))


def deque_constructor(constructor, node):
    """Constructor for ndarray"""
    return deque(constructor.construct_sequence(node))


def lambda_representer(representer, data):
    """Represent a lambda function by serializing its code."""
    serialized_lambda = base64.b64encode(dumps(data)).decode("utf-8")
    return representer.represent_scalar("!lambda", serialized_lambda)


def lambda_constructor(constructor, node):
    """Reconstruct a lambda function from the serialized data."""
    # Decode the base64-encoded string and load the lambda function
    serialized_lambda = base64.b64decode(node.value)
    return loads(serialized_lambda)  # noqa: S301


yaml = YAML(typ="unsafe")
yaml.register_class(UUID)
yaml.representer.add_representer(np.ndarray, ndarray_representer)
yaml.constructor.add_constructor("!ndarray", ndarray_constructor)
yaml.representer.add_representer(deque, deque_representer)
yaml.constructor.add_constructor("!deque", deque_constructor)
yaml.representer.add_representer(types.LambdaType, lambda_representer)
yaml.constructor.add_constructor("!lambda", lambda_constructor)



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Qblox Acquisitions Builder"""

from qililab.result.qblox_results.bins_data import BinsData
from qililab.result.qblox_results.qblox_bins_acquisitions import QbloxBinsAcquisitions
from qililab.result.qblox_results.qblox_scope_acquisitions import QbloxScopeAcquisitions
from qililab.result.qblox_results.scope_data import ScopeData


class QbloxAcquisitionsBuilder:
    """Qblox Acquisitions Results Builder"""

    @staticmethod
    def get_scope(integration_lengths: list[int], qblox_raw_results: list[dict]) -> QbloxScopeAcquisitions | None:
        """Cast dictionaries to their corresponding class."""
        sequencer_scope = QbloxAcquisitionsBuilder._get_sequencer_with_scope_data_available(qblox_raw_results)
        if sequencer_scope is not None:
            scope_data = qblox_raw_results[sequencer_scope]["scope"]
            return QbloxScopeAcquisitions(
                pulse_length=integration_lengths[sequencer_scope], scope=ScopeData(**scope_data)
            )
        return None

    @staticmethod
    def get_bins(integration_lengths: list[int], qblox_raw_results: list[dict]) -> QbloxBinsAcquisitions:
        """Cast dictionaries to their corresponding class."""
        bins_data = [sequencer_acq["bins"] for sequencer_acq in qblox_raw_results]
        return QbloxBinsAcquisitions(
            integration_lengths=integration_lengths,
            bins=[BinsData(**sequencer_bins_data) for sequencer_bins_data in bins_data],
        )

    @staticmethod
    def _get_sequencer_with_scope_data_available(qblox_raw_results: list[dict]) -> int | None:
        """Returns the sequencer with scope data available.

        Returns:
            int | None: Number of the sequencer with scope data. None if there is no scope data available in any sequencer.
        """
        return next(
            (
                sequencer
                for sequencer, acquitision in enumerate(qblox_raw_results)
                if any(acquitision["scope"]["path0"]["data"])
            ),
            None,
        )



import io
import types
from collections import deque
from uuid import UUID

import numpy as np

from qililab.yaml import yaml


def test_ndarray_serialization():
    """Test serialization and deserialization of numpy ndarray."""
    # Create a sample ndarray
    original_array = np.array([[1, 2, 3], [4, 5, 6]], dtype=np.float64)

    # Dump the array to YAML
    stream = io.StringIO()
    yaml.dump(original_array, stream)
    yaml_str = stream.getvalue()

    # Load the array back from YAML
    loaded_array = yaml.load(yaml_str)

    # Verify that the original and loaded arrays are equal
    np.testing.assert_array_equal(original_array, loaded_array)


def test_deque_serialization():
    """Test serialization and deserialization of deque."""
    # Create a sample deque
    original_deque = deque([1, 2, 3, 4, 5])

    # Dump the deque to YAML
    stream = io.StringIO()
    yaml.dump(original_deque, stream)
    yaml_str = stream.getvalue()

    # Load the deque back from YAML
    loaded_deque = yaml.load(yaml_str)

    # Verify that the original and loaded deques are equal
    assert original_deque == loaded_deque
    assert isinstance(loaded_deque, deque)


def test_lambda_serialization():
    """Test serialization and deserialization of a lambda function."""

    def another_function(y):
        return y + 1

    # Define a sample lambda function
    def original_lambda(x, y):
        return x + another_function(y)

    # Dump the lambda function to YAML
    stream = io.StringIO()
    yaml.dump(original_lambda, stream)
    yaml_str = stream.getvalue()

    # Load the lambda function back from YAML
    loaded_lambda = yaml.load(yaml_str)

    # Verify that the loaded object is a function
    assert isinstance(loaded_lambda, types.FunctionType)

    # Test that the loaded lambda behaves the same as the original
    x, y = 5, 10
    assert original_lambda(x, y) == loaded_lambda(x, y)


def test_uuid_serialization():
    """Test serialization and deserialization of UUID."""
    # Create a sample UUID
    original_uuid = UUID("12345678123456781234567812345678")

    # Dump the UUID to YAML
    stream = io.StringIO()
    yaml.dump(original_uuid, stream)
    yaml_str = stream.getvalue()

    # Load the UUID back from YAML
    loaded_uuid = yaml.load(yaml_str)

    # Verify that the original and loaded UUIDs are equal
    assert original_uuid == loaded_uuid
    assert isinstance(loaded_uuid, UUID)



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Class to interface with the voltage source Qblox D5a
"""

from dataclasses import dataclass
from time import sleep
from typing import Any

from qililab.config import logger
from qililab.instruments.decorators import check_device_initialized, log_set_parameter
from qililab.instruments.instrument import ParameterNotFound
from qililab.instruments.utils import InstrumentFactory
from qililab.instruments.voltage_source import VoltageSource
from qililab.typings import ChannelID, InstrumentName, Parameter, ParameterValue
from qililab.typings import QbloxD5a as QbloxD5aDriver


@InstrumentFactory.register
class QbloxD5a(VoltageSource):
    """Qblox D5a class

    Args:
        name (InstrumentName): name of the instrument
        device (Qblox_D5a): Instance of the qcodes D5a class.
        settings (QbloxD5aSettings): Settings of the instrument.
    """

    name = InstrumentName.QBLOX_D5A

    @dataclass
    class QbloxD5aSettings(VoltageSource.VoltageSourceSettings):
        """Contains the settings of a specific signal generator."""

    settings: QbloxD5aSettings
    device: QbloxD5aDriver

    def dac(self, dac_index: int):
        """get channel associated to the specific dac

        Args:
            dac_index (int): channel index

        Returns:
            _type_: _description_
        """
        return getattr(self.device, f"dac{dac_index}")

    @log_set_parameter
    def set_parameter(self, parameter: Parameter, value: ParameterValue, channel_id: ChannelID | None = None):
        """Set Qblox instrument calibration settings."""

        if channel_id is None:
            raise ValueError(f"channel not specified to update instrument {self.name.value}")

        channel_id = int(channel_id)
        if channel_id > 15:
            raise ValueError(
                f"the specified dac index:{channel_id} is out of range."
                + " Number of dacs is 16 -> maximum channel_id should be 15."
            )

        channel = self.dac(dac_index=channel_id) if self.is_device_active() else None

        if parameter == Parameter.VOLTAGE:
            self._set_voltage(value=value, channel_id=channel_id, channel=channel)
            return
        if parameter == Parameter.SPAN:
            self._set_span(value=value, channel_id=channel_id, channel=channel)
            return
        if parameter == Parameter.RAMPING_ENABLED:
            self._set_ramping_enabled(value=value, channel_id=channel_id, channel=channel)
            return
        if parameter == Parameter.RAMPING_RATE:
            self._set_ramping_rate(value=value, channel_id=channel_id, channel=channel)
            return
        raise ParameterNotFound(self, parameter)

    def get_parameter(self, parameter: Parameter, channel_id: ChannelID | None = None):
        """Get instrument parameter.

        Args:
            parameter (Parameter): Name of the parameter to get.
            channel_id (int | None): Channel identifier of the parameter to update.
        """
        if channel_id is None:
            raise ValueError(f"channel not specified to update instrument {self.name.value}")

        channel_id = int(channel_id)
        if channel_id > 15:
            raise ValueError(
                f"the specified dac index:{channel_id} is out of range."
                + " Number of dacs is 16 -> maximum channel_id should be 15."
            )
        if hasattr(self.settings, parameter.value):
            return getattr(self.settings, parameter.value)[channel_id]
        raise ParameterNotFound(self, parameter)

    def _set_voltage(self, value: float | str | bool, channel_id: int, channel: Any):
        """Set the voltage"""
        self.settings.voltage[channel_id] = float(value)
        if self.is_device_active():
            channel.voltage(self.voltage[channel_id])

    def _set_span(self, value: float | str | bool, channel_id: int, channel: Any):
        """Set the span"""
        self.settings.span[channel_id] = str(value)
        if self.is_device_active():
            channel.span(self.span[channel_id])

    def _set_ramping_enabled(self, value: float | str | bool, channel_id: int, channel: Any):
        """Set the ramping_enabled"""
        self.settings.ramping_enabled[channel_id] = bool(value)
        if self.is_device_active():
            channel.ramping_enabled(self.ramping_enabled[channel_id])

    def _set_ramping_rate(self, value: float | str | bool, channel_id: int, channel: Any):
        """Set the ramp_rate"""
        self.settings.ramp_rate[channel_id] = float(value)
        if self.is_device_active():
            channel.ramp_rate(self.ramp_rate[channel_id])

    @check_device_initialized
    def initial_setup(self):
        """performs an initial setup."""
        for channel_id in self.dacs:
            index = self.dacs.index(channel_id)
            channel = self.dac(dac_index=channel_id)
            channel.ramping_enabled(self.ramping_enabled[index])
            channel.ramp_rate(self.ramp_rate[index])
            channel.span(self.span[index])
            channel.voltage(self.voltage[index])
            logger.debug("SPI voltage set to %f", channel.voltage())
            while channel.is_ramping():
                sleep(0.1)

    @check_device_initialized
    def turn_on(self):
        """Dummy method."""

    @check_device_initialized
    def turn_off(self):
        """Stop outputing voltage."""
        for dac_index in self.settings.dacs:
            channel = self.dac(dac_index=dac_index)
            channel.voltage(0)
            logger.debug("Dac%d voltage resetted to  %f", dac_index, channel.voltage())

    @check_device_initialized
    def reset(self):
        """Reset instrument."""
        for dac_index in self.settings.dacs:
            channel = self.dac(dac_index=dac_index)
            channel.voltage(0)



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""PulseSequence class."""

from dataclasses import dataclass, field

from qililab.constants import PULSESCHEDULES
from qililab.pulse.pulse_bus_schedule import PulseBusSchedule
from qililab.pulse.pulse_event import PulseEvent


@dataclass
class PulseSchedule:
    """Class containing a list of :class:`PulseBusSchedule` objects. It is the pulsed representation of a Qibo circuit.

    This class will receive a list of :class:`PulseBusSchedule` objects as argument. The class allows several operations to this
    list of PulseSequence objects and it is the class that allows to transpile pulse sequences into programs that run on quantum
    hardware control instruments.

    Args:
        elements (list[PulseBusSchedule]): List of :class:`PulseBusSchedule`.

    Examples:
        Imagine you want to create a PulseSchedule instance that will contain two :class:`PulseBusSchedule` objects, one for driving a qubit
        and the other one to do readout on the same qubit.

        To do so

        .. code-block:: python3

            drag_pulse = Pulse(
                amplitude=1, phase=0.5, duration=200, frequency=1e9, pulse_shape=Drag(num_sigmas=4, drag_coefficient=0.5)
            )
            readout_pulse = Pulse(amplitude=1, phase=0.5, duration=1500, frequency=1e9, pulse_shape=Rectangular())
            drag_pulse_event = PulseEvent(pulse=drag_pulse, start_time=0)
            readout_pulse_event = PulseEvent(pulse=readout_pulse, start_time=200, qubit=0)

            drive_schedule = PulseBusSchedule(timeline=[drag_pulse_event], bus_alias="drive_q0")
            readout_schedule = PulseBusSchedule(timeline=[pulse_event], bus_alias="feedline_input")
            list_pulse_bus_schedules = [drive_schedule, readout_schedule]

            pulse_schedule = PulseSchedule(list_pulse_bus_schedules)

        You can also create a pulse schedule by creating an instance of PulseSchedule class, with no pulse bus schedules passed
        as arguments, and dynamically adding pulses to the PulseSchedule class, which will then create a :class:`PulseBusSchedule` for
        each bus the pulses will be using.

        To do so, you can first create an instance of the PulseSchedule class, create each of the two pulses by creating two
        instances of the Pulse class, and then use the `add_event()` method of the PulseSchedule class to add the two pulses, wrapped
        by :class:`PulseEvent` class, to the schedule:

        .. code-block:: python3

            pulse_schedule = PulseSchedule()
            drag_pulse = Pulse(
                amplitude=1, phase=0.5, duration=200, frequency=1e9, pulse_shape=Drag(num_sigmas=4, drag_coefficient=0.5)
            )
            readout_pulse = Pulse(amplitude=1, phase=0.5, duration=1500, frequency=1e9, pulse_shape=Rectangular())
            pulse_schedule.add_event(PulseEvent(pulse=drag_pulse, start_time=0), bus_alias="drive_q0", delay=0)
            pulse_schedule.add_event(
                PulseEvent(pulse=readout_pulse, start_time=200, qubit=0), bus_alias="feedline_input", delay=0
            )

        It is possible to serialize a PulseSchedule object as a dictionary. To do so you can use the `to_dict()` method:

        .. code-block:: python3

            pulse_schedule = PulseSchedule()
            drag_pulse = Pulse(
                amplitude=1, phase=0.5, duration=200, frequency=1e9, pulse_shape=Drag(num_sigmas=4, drag_coefficient=0.5)
            )
            readout_pulse = Pulse(amplitude=1, phase=0.5, duration=1500, frequency=1e9, pulse_shape=Rectangular())
            pulse_schedule.add_event(PulseEvent(pulse=drag_pulse, start_time=0), bus_alias="drive_q0", delay=0)
            pulse_schedule.add_event(
                PulseEvent(pulse=readout_pulse, start_time=200, qubit=0), bus_alias="feedline_input", delay=0
            )

            pulse_schedule_dict = pulse_schedule.to_dict()

        It is also possible to create a PulseSchedule object from a dictionary that contains the serialized description of the
        object. To do so you can use the `from_dict()` method:

        .. code-block:: python3

            pulse_schedule = PulseSchedule()
            drag_pulse = Pulse(
                amplitude=1, phase=0.5, duration=200, frequency=1e9, pulse_shape=Drag(num_sigmas=4, drag_coefficient=0.5)
            )
            readout_pulse = Pulse(amplitude=1, phase=0.5, duration=1500, frequency=1e9, pulse_shape=Rectangular())
            pulse_schedule.add_event(PulseEvent(pulse=drag_pulse, start_time=0), bus_alias="drive_q0", delay=0)
            pulse_schedule.add_event(
                PulseEvent(pulse=readout_pulse, start_time=200, qubit=0), bus_alias="feedline_input", delay=0
            )

            pulse_schedule_dict = pulse_schedule.to_dict()

            pulse_schedule = PulseSchedule.from_dict(pulse_schedule_dict)
    """

    elements: list[PulseBusSchedule] = field(default_factory=list)  #: List of pulse bus schedules.

    def add_event(self, pulse_event: PulseEvent, bus_alias: str, delay: int):
        """Adds pulse event to the list of pulse bus schedules.

        This functions receives a :class:`PulseEvent` object, a bus alias and a delay parameter, and checks whether
        there is already a PulseBusSchedule for the given bus, adding the pulse event to the :class:`PulseBusSchedule`.
        If there is not a :class:`PulseBusSchedule` for that bus, it creates a new one passing the pulse event and
        bus alias as parameters, and adds this new instance to the list of :class:`PulseBusSchedule`.

        Args:
            pulse_event (PulseEvent): :class:`PulseEvent` object.
            bus_alias (str): Alias of the bus targeted by the pulse event.
            delay (int): Delay (in ns) of the pulse event. This delay is added at the beginning of the :class:`PulseEvent`.
        """
        pulse_event.start_time += delay
        for pulse_sequence in self.elements:
            if bus_alias == pulse_sequence.bus_alias:
                pulse_sequence.add_event(pulse_event=pulse_event)
                return
        self.elements.append(PulseBusSchedule(timeline=[pulse_event], bus_alias=bus_alias))

    def create_schedule(self, bus_alias: str):
        """Creates an empty `PulseBusSchedule` that targets the given bus.

        If the schedule already exists, nothing is done.

        Args:
            bus_alias (str): Target bus alias of the schedule to create.
        """
        aliases = {schedule.bus_alias for schedule in self.elements}
        if bus_alias not in aliases:
            self.elements.append(PulseBusSchedule(bus_alias=bus_alias))

    def to_dict(self):
        """Returns dictionary representation of the class.

        Returns:
            dict: Dictionary representation of the class.
        """
        return {PULSESCHEDULES.ELEMENTS: [pulse_sequence.to_dict() for pulse_sequence in self.elements]}

    @classmethod
    def from_dict(cls, dictionary: dict):
        """Builds PulseSequence instance from dictionary.

        Args:
            dictionary (dict): Dictionary description of the class.

        Returns:
            PulseSequence: Class instance.
        """
        elements = [PulseBusSchedule.from_dict(dictionary=settings) for settings in dictionary[PULSESCHEDULES.ELEMENTS]]

        return cls(elements=elements)

    def __iter__(self):
        """Redirects __iter__ magic method to elements."""
        return self.elements.__iter__()

    def __len__(self):
        """Redirects __len__ magic method to elements."""
        return len(self.elements)



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
This module contains all the methods and the class used to define a platform, which is a representation
of a laboratory.

Platform Class
~~~~~~~~~~~~~~~~

.. currentmodule:: qililab.platform

.. autosummary::
    :toctree: api

    ~Platform

"""

from .components import Bus, BusElement, Buses
from .platform import Platform

__all__ = ["Bus", "BusElement", "Buses", "Platform"]



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Waveform class."""

from dataclasses import dataclass, field

import numpy as np


@dataclass
class Waveforms:
    """Waveform class that containg the I and Q modulated waveforms."""

    i: np.ndarray = field(default_factory=lambda: np.array([]))
    q: np.ndarray = field(default_factory=lambda: np.array([]))

    def add(self, imod: np.ndarray, qmod: np.ndarray):
        """Add i and q arrays to the waveforms.

        Args:
            imod (np.ndarray): I modulated waveform to add.
            qmod (np.ndarray): Q modulated waveform to add.
        """
        self.i = np.append(self.i, imod)
        self.q = np.append(self.q, qmod)

    @property
    def values(self):
        """Return the waveform i and q values.

        Returns:
            np.ndarray: Array containing the i and q waveform values.
        """
        return np.array([self.i, self.q])

    def __add__(self, other):
        """Add two Waveform objects."""
        if not isinstance(other, Waveforms):
            raise NotImplementedError
        self.i = np.append(self.i, other.i)
        self.q = np.append(self.q, other.q)
        return self

    def __len__(self):
        """Length of the object."""
        if len(self.i) != len(self.q):
            raise ValueError("Both I and Q waveforms must have the same length.")
        return len(self.i)



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""__init__.py"""

from .enums import (
    AcquireTriggerMode,
    AcquisitionName,
    ConnectionName,
    GateName,
    Instrument,
    InstrumentControllerName,
    InstrumentName,
    IntegrationMode,
    Parameter,
    PulseDistortionName,
    PulseShapeName,
    ReferenceClock,
    ResultName,
)
from .factory_element import FactoryElement
from .instruments import (
    Cluster,
    Device,
    Keithley2600Driver,
    MiniCircuitsDriver,
    QbloxD5a,
    QbloxS4g,
    QcmQrm,
    QDevilQDac2,
    QMMDriver,
    RohdeSchwarzSGS100A,
    YokogawaGS200,
)
from .type_aliases import ChannelID, ParameterValue

__all__ = [
    "AcquireTriggerMode",
    "AcquisitionName",
    "ChannelID",
    "Cluster",
    "ConnectionName",
    "Device",
    "FactoryElement",
    "GateName",
    "Instrument",
    "InstrumentControllerName",
    "InstrumentName",
    "IntegrationMode",
    "Keithley2600Driver",
    "MiniCircuitsDriver",
    "Parameter",
    "ParameterValue",
    "PulseDistortionName",
    "PulseShapeName",
    "QDevilQDac2",
    "QMMDriver",
    "QbloxD5a",
    "QbloxS4g",
    "QcmQrm",
    "ReferenceClock",
    "ResultName",
    "RohdeSchwarzSGS100A",
    "YokogawaGS200",
]



from copy import deepcopy
import re
from unittest.mock import patch
import numpy as np
from qibo import Circuit, gates
import pytest

from qililab.digital.circuit_optimizer import CircuitOptimizer
from qililab.digital.native_gates import Drag
from tests.data import Galadriel
from tests.test_utils import build_platform


class TestCircuitOptimizerIntegration:
    """Tests for the circuit optimizer class, with integration tests."""

    def test_run_gate_cancelation(self):
        """Test run gate cancelation."""
        # Create a circuit with two gates that cancel each other.
        circuit = Circuit(5)

        # pairs that cancels:
        circuit.add(gates.H(0))
        circuit.add(gates.H(0))

        # From here only the X(4) will cancel with the X(4) at the end.
        circuit.add(gates.CNOT(2,3)) # 1
        circuit.add(gates.X(4))
        circuit.add(gates.H(3)) # 2

        # The 0-1 and 1-4 CNOTs shold cancel each other.
        circuit.add(gates.CNOT(1,4))
        circuit.add(gates.CNOT(0,1))
        circuit.add(Drag(3, theta=2*np.pi, phase=np.pi)) # 3
        circuit.add(gates.CNOT(0,1))
        circuit.add(gates.CNOT(1,4))

        circuit.add(gates.H(3)) # 4
        circuit.add(gates.X(4))
        circuit.add(gates.CNOT(2,3)) # 5


        # Run Optimize the circuit.
        optimizer = CircuitOptimizer(None)
        optimized_gates_hermitian_cancel = optimizer.cancel_pairs_of_hermitian_gates(circuit.queue)
        optimized_gates_complete = optimizer.optimize_gates(circuit.queue)

        # Check that the circuit is optimized
        assert len(optimized_gates_hermitian_cancel) == 5
        assert len(optimized_gates_complete) == 4
        # Check name attribute:
        assert [gate.name for gate in optimized_gates_hermitian_cancel] == ["cx", "h", "drag", "h", "cx"]
        assert [gate.name for gate in optimized_gates_complete] == ["h", "drag", "h", "cx"]
        # CHeck the type of the gates:
        assert [type(gate).__name__ for gate in optimized_gates_hermitian_cancel] == ["CNOT", "H", "Drag", "H", "CNOT"]
        assert [type(gate).__name__ for gate in optimized_gates_complete] == ["H", "Drag", "H", "CNOT"]
        # Assert the initial arguments:
        assert [gate.init_args for gate in optimized_gates_hermitian_cancel] == [[2,3], [3], [3], [3], [2,3]]
        assert [gate.init_args for gate in optimized_gates_complete] == [[3], [3], [3], [2,3]]
        assert [gate.init_kwargs for gate in optimized_gates_hermitian_cancel] == [{}, {}, {"theta": 2*np.pi, "phase": np.pi, "trainable": True}, {}, {}]
        assert [gate.init_kwargs for gate in optimized_gates_complete] == [{}, {"theta": 2*np.pi, "phase": np.pi, "trainable": True}, {}, {}]


class TestCircuitOptimizerUnit:
    """Tests for the circuit optimizer class, with Unit test."""

    @patch("qililab.digital.circuit_optimizer.CircuitOptimizer.cancel_pairs_of_hermitian_gates", return_value=[gates.CZ(0, 1), Drag(0, theta=np.pi, phase=np.pi / 2)])
    @patch("qililab.digital.circuit_optimizer.CircuitOptimizer.remove_redundant_start_controlled_gates", return_value=[Drag(0, theta=np.pi, phase=np.pi / 2)])
    def test_run_gate_cancellations(self, mock_redundant_gates, mock_cancellation):
        """Test optimize transpilation."""
        circuit = Circuit(2)
        circuit.add(gates.RZ(0, theta=np.pi / 2))
        circuit.add(gates.CZ(0, 1))
        circuit.add(Drag(0, theta=np.pi, phase=np.pi / 2))

        optimizer = CircuitOptimizer(None)
        optimized_gates_complete = optimizer.optimize_gates(circuit.queue)
        mock_cancellation.assert_called_once_with(circuit.queue)
        mock_redundant_gates.assert_called_once_with(mock_cancellation.return_value)

        # Assert remaining gates:
        assert len(optimized_gates_complete) == 1
        assert [gate.name for gate in optimized_gates_complete] == ["drag"]
        assert [type(gate).__name__ for gate in optimized_gates_complete] == ["Drag"]


    @patch("qililab.digital.native_gates._GateHandler.create_qibo_gates_from_gates_info", return_value=Circuit(5).queue)
    @patch("qililab.digital.circuit_optimizer.CircuitOptimizer._sweep_circuit_cancelling_pairs_of_hermitian_gates", return_value=[("CZ", [0, 1], {}), ("Drag", [0], {"theta": np.pi, "phase": np.pi / 2})])
    @patch("qililab.digital.native_gates._GateHandler.get_circuit_gates_info", return_value=[("CZ", [0, 1], {}), ("Drag", [0], {"theta": np.pi, "phase": np.pi / 2})])
    def test_cancel_pairs_of_hermitian_gates(self, mock_get_circuit_gates_info, mock_sweep_circuit, mock_create_circuit):
        """Test run gate cancellations with mocks."""
        circuit = Circuit(2)
        circuit.add(gates.RZ(0, theta=np.pi / 2))
        circuit.add(gates.CZ(0, 1))
        circuit.add(Drag(0, theta=np.pi, phase=np.pi / 2))

        optimizer = CircuitOptimizer(None)
        _ = optimizer.cancel_pairs_of_hermitian_gates(circuit.queue)

        mock_get_circuit_gates_info.assert_called_once_with(circuit.queue)
        mock_sweep_circuit.assert_called_once_with([("CZ", [0, 1], {}), ("Drag", [0], {"theta": np.pi, "phase": np.pi / 2})])
        mock_create_circuit.assert_called_once_with([("CZ", [0, 1], {}), ("Drag", [0], {"theta": np.pi, "phase": np.pi / 2})])


    def test_sweep_circuit_cancelling_pairs_of_hermitian_gates(self):
        """Test sweep circuit cancelling pairs of hermitian gates."""
        circuit_gates = [("X", [0], {}), ("X", [0], {}), ("H", [1], {}), ("H", [1], {})]
        output_circuit_gates = CircuitOptimizer._sweep_circuit_cancelling_pairs_of_hermitian_gates(circuit_gates)

        assert output_circuit_gates == []


    def test_merge_consecutive_drags_diff_qubits(self):
        """Test merge drag gates, with incorrect input."""
        drag_1 = Drag(0, theta=np.pi, phase=np.pi / 2)
        drag_2 = Drag(1, theta=np.pi, phase=np.pi / 2)

        optimizer = CircuitOptimizer(None)
        with pytest.raises(ValueError, match=re.escape("Cannot merge Drag gates acting on different qubits.")):
            _ = optimizer.merge_consecutive_drags(drag_1, drag_2)

    def test_merge_consecutive_drags_same_phis(self):
        """Test merge drag gates."""
        drag_1 = Drag(0, theta=np.pi/2-0.1, phase=np.pi / 2)
        drag_2 = Drag(0, theta=np.pi/2, phase=np.pi / 2)

        optimizer = CircuitOptimizer(None)
        final_drag = optimizer.merge_consecutive_drags(drag_1, drag_2)

        assert isinstance(final_drag, Drag)
        assert final_drag.parameters == (np.pi-0.1, np.pi / 2)

    def test_merge_consecutive_drags_opposite_phis(self):
        """Test merge drag gates."""
        drag_1 = Drag(0, theta=np.pi/2+0.1, phase=(np.pi / 2)+0.1)
        drag_2 = Drag(0, theta=np.pi/2, phase=-(np.pi / 2)+0.1)

        optimizer = CircuitOptimizer(None)
        final_drag = optimizer.merge_consecutive_drags(drag_1, drag_2)

        assert isinstance(final_drag, Drag)
        assert np.isclose(final_drag.parameters, (0.1, (np.pi / 2)+0.1)).all()

    def test_merge_consecutive_drags_diff_phis(self):
        """Test merge drag gates."""
        drag_1 = Drag(0, theta=np.pi, phase=np.pi / 2)
        drag_2 = Drag(0, theta=np.pi, phase=np.pi)

        optimizer = CircuitOptimizer(None)
        final_drag = optimizer.merge_consecutive_drags(drag_1, drag_2)

        assert final_drag is None


    def test_bunch_drag_gates_only_same_phis(self):
        """Test bunch drag gates."""
        circuit = Circuit(2)
        circuit.add(Drag(0, theta=np.pi, phase=np.pi / 2))
        circuit.add(Drag(0, theta=np.pi, phase=np.pi / 2))
        circuit.add(Drag(1, theta=np.pi, phase=np.pi / 2))
        circuit.add(Drag(1, theta=np.pi, phase=np.pi / 2))
        circuit.add(Drag(0, theta=np.pi, phase=np.pi / 2))
        circuit.add(Drag(0, theta=np.pi, phase=np.pi))
        circuit.add(gates.RX(1, theta=np.pi / 2))
        circuit.add(Drag(1, theta=np.pi, phase=np.pi / 2))


        optimizer = CircuitOptimizer(None)
        gate_list = optimizer.bunch_drag_gates(circuit.queue)

        assert len(gate_list) == 5

        assert isinstance(gate_list[0], Drag)
        assert gate_list[0].parameters == (np.pi, np.pi / 2)

        assert isinstance(gate_list[1], Drag)
        assert gate_list[1].parameters == (0, np.pi / 2)

        assert isinstance(gate_list[2], Drag)
        assert gate_list[2].parameters == (np.pi, np.pi)

        assert isinstance(gate_list[3], gates.RX)

        assert isinstance(gate_list[4], Drag)
        assert gate_list[4].parameters == (np.pi, np.pi / 2)

    def test_bunch_drag_gates_diff_phis(self):
        """Test bunch drag gates."""
        circuit = Circuit(2)
        circuit.add(Drag(0, theta=np.pi, phase=np.pi / 2))
        circuit.add(Drag(0, theta=np.pi, phase=np.pi))

        optimizer = CircuitOptimizer(None)
        gate_list = optimizer.bunch_drag_gates(circuit.queue)

        assert gate_list == circuit.queue

    def test_delete_gates_with_no_amplitude(self):
        """Test delete gates with no amplitude."""
        circuit = Circuit(2)
        circuit.add(Drag(0, theta=0, phase=np.pi / 2))
        circuit.add(Drag(0, theta=np.pi, phase=np.pi / 2))

        gate_list = CircuitOptimizer.delete_gates_with_no_amplitude(circuit.queue)

        assert len(gate_list) == 1
        assert isinstance(gate_list[0], Drag)
        assert gate_list[0].parameters == (np.pi, np.pi / 2)

    def test_normalize_angles_of_drags(self):
        """Test normalize angles of drags."""
        circuit = Circuit(2)
        circuit.add(Drag(0, theta=3 * np.pi, phase=np.pi / 2))
        circuit.add(Drag(0, theta=2 * np.pi, phase=np.pi / 2))
        circuit.add(Drag(0, theta=np.pi, phase=2*np.pi))
        circuit.add(Drag(0, theta=np.pi, phase=np.pi / 2))
        circuit.add(Drag(0, theta=np.pi, phase=np.pi / 2))
        circuit.add(Drag(0, theta=np.pi, phase=np.pi / 2))

        gate_list = CircuitOptimizer.normalize_angles_of_drags(circuit.queue)

        assert len(gate_list) == 6
        assert [gate.parameters for gate in gate_list] == [
            (np.pi, np.pi / 2),
            (0, np.pi / 2),
            (np.pi, 0),
            (np.pi, np.pi / 2),
            (np.pi, np.pi / 2),
            (np.pi, np.pi / 2),
        ]

        ###########

    @pytest.mark.parametrize(
        "wire_names", [None, [0, 1]]
    )
    def test_make_gate_None_in_queue(self, wire_names):
        """Test make gate None in queue."""
        circuit = Circuit(2)
        circuit.add(gates.SWAP(0, 1))
        circuit.add(gates.X(0))

        queue = circuit.queue
        wire_names_was_None = not deepcopy(wire_names)

        CircuitOptimizer._make_gate_None_in_queue(queue[0], 0, queue, wire_names)

        assert queue[0] is None
        if wire_names_was_None:
            assert wire_names is None
        else:
            assert wire_names == [1, 0]


    def test_make_gate_None_in_queue_no_wire_names(self):
        """Test make gate None in queue."""
        circuit = Circuit(2)
        circuit.add(gates.SWAP(0, 1))
        circuit.add(gates.X(0))

        queue = circuit.queue
        wire_names = None

        CircuitOptimizer._make_gate_None_in_queue(queue[0], 0, queue, wire_names)

        assert queue[0] is None
        assert wire_names is None

    def test_remove_redundant_start_controlled_gates(self):
        """Test remove redundant start controlled gates."""
        circuit = Circuit(3)
        circuit.add(gates.CNOT(0, 1))
        circuit.add(gates.CNOT(1, 2))
        circuit.add(gates.H(0))
        circuit.add(gates.CNOT(0, 1))

        optimizer = CircuitOptimizer(None)
        optimized_circuit = optimizer.remove_redundant_start_controlled_gates(circuit)

        assert len(optimized_circuit.queue) == 2
        assert [gate.name for gate in optimized_circuit.queue] == ["h", "cx"]

    def test_add_phases_from_RZs_and_CZs_to_drags(self):
        """Test add phases from RZs and CZs to drags."""
        circuit = Circuit(2)
        circuit.add(gates.RZ(0, theta=np.pi / 2))
        circuit.add(gates.CZ(0, 1))
        circuit.add(Drag(0, theta=np.pi, phase=np.pi / 2)) # Should end up, with the phase cancelled by RZ and then a correction added by CZ.

        platform = build_platform(runcard=Galadriel.runcard)
        optimizer = CircuitOptimizer(platform.digital_compilation_settings)
        optimized_gates = optimizer.add_phases_from_RZs_and_CZs_to_drags(circuit.queue, 2)

        assert len(optimized_gates) == 2
        assert isinstance(optimized_gates[1], Drag)
        correction_from_CZ_in_qubit_0= platform.digital_compilation_settings.gates["CZ(0,1)"][0].pulse.options["q0_phase_correction"]
        assert np.isclose(optimized_gates[1].parameters, (np.pi, -correction_from_CZ_in_qubit_0)).all

    def test_optimize_transpiled_gates_all_cancels(self):
        """Test optimize transpiled gates."""
        circuit = Circuit(2)
        circuit.add(Drag(0, theta=np.pi, phase=np.pi / 2))
        circuit.add(Drag(0, theta=np.pi, phase=np.pi / 2))
        circuit.add(Drag(1, theta=0, phase=np.pi / 2))

        optimizer = CircuitOptimizer(None)
        optimized_gates = optimizer.optimize_transpiled_gates(circuit.queue)

        assert len(optimized_gates) == 0

    def test_optimize_transpiled_gates(self):
        """Test optimize transpiled gates."""
        circuit = Circuit(2)
        circuit.add(Drag(0, theta=np.pi, phase=np.pi / 2))
        circuit.add(Drag(0, theta=np.pi/2, phase=np.pi / 2))
        circuit.add(Drag(1, theta=np.pi, phase=np.pi / 2))

        optimizer = CircuitOptimizer(None)
        optimized_gates = optimizer.optimize_transpiled_gates(circuit.queue)

        assert len(optimized_gates) == 2
        assert [gate.parameters for gate in optimized_gates] == [(-np.pi /2, np.pi / 2), (np.pi, np.pi / 2)]



"""Tests for the SGS100A class."""
from unittest.mock import MagicMock

import pytest

from qililab.instruments.instrument import ParameterNotFound
from qililab.instruments.yokogawa.gs200 import GS200
from qililab.typings.enums import Parameter, SourceMode


@pytest.fixture(name="yokogawa_gs200")
def fixture_yokogawa_gs200():
    """Return connected instance of GS200 class"""
    yokogawa_gs200_current = GS200({
        "alias": "yokogawa_current",
        Parameter.SOURCE_MODE.value: "current",
        Parameter.CURRENT.value: [0.5],
        Parameter.VOLTAGE.value: [0.0],
        Parameter.SPAN.value: ["200mA"],
        Parameter.RAMPING_ENABLED.value: [True],
        Parameter.RAMPING_RATE.value: [0.01],
        "dacs": [0],
    })
    yokogawa_gs200_current.device = MagicMock()
    yokogawa_gs200_current.device.mock_add_spec(["current", "voltage", "source_mode", "current_range", "voltage_range", "ramp_current", "ramp_voltage", "on", "off"])
    return yokogawa_gs200_current


@pytest.fixture(name="yokogawa_gs200_voltage")
def fixture_yokogawa_gs200_voltage():
    """Return connected instance of GS200 class"""
    yokogawa_gs200_voltage = GS200({
        "alias": "yokogawa_current",
        Parameter.SOURCE_MODE.value: "voltage",
        Parameter.CURRENT.value: [0.0],
        Parameter.VOLTAGE.value: [0.5],
        Parameter.SPAN.value: ["1V"],
        Parameter.RAMPING_ENABLED.value: [True],
        Parameter.RAMPING_RATE.value: [0.01],
        "dacs": [0],
    })
    yokogawa_gs200_voltage.device = MagicMock()
    yokogawa_gs200_voltage.device.mock_add_spec(["current", "voltage", "source_mode", "current_range", "voltage_range", "ramp_current", "ramp_voltage", "on", "off"])
    return yokogawa_gs200_voltage


class TestYokogawaGS200:
    """Unit tests checking the GS200 attributes and methods"""

    @pytest.mark.parametrize(
        "parameter, expected_value",
        [
            (Parameter.CURRENT, 0.5),
            (Parameter.VOLTAGE, 0.0),
            (Parameter.RAMPING_ENABLED, True),
            (Parameter.RAMPING_RATE, 0.01),
            (Parameter.SOURCE_MODE, SourceMode.CURRENT),
            (Parameter.SPAN, "200mA"),
        ],
    )
    def test_get_parameter_method(self, parameter: Parameter, expected_value, yokogawa_gs200: GS200):
        """Test the set_parameter method with float value"""
        value = yokogawa_gs200.get_parameter(parameter)
        assert value == expected_value

    @pytest.mark.parametrize(
        "parameter, expected_value",
        [
            (Parameter.CURRENT, 0.0),
            (Parameter.VOLTAGE, 0.5),
            (Parameter.RAMPING_ENABLED, True),
            (Parameter.RAMPING_RATE, 0.01),
            (Parameter.SOURCE_MODE, SourceMode.VOLTAGE),
            (Parameter.SPAN, "1V"),
        ],
    )
    def test_get_parameter_method_voltage(self, parameter: Parameter, expected_value, yokogawa_gs200_voltage: GS200):
        """Test the set_parameter method with float value"""
        value = yokogawa_gs200_voltage.get_parameter(parameter)
        assert value == expected_value

    @pytest.mark.parametrize("parameter", [Parameter.MAX_CURRENT, Parameter.GAIN])
    def test_get_parameter_method_raises_exception(self, parameter: Parameter, yokogawa_gs200: GS200):
        """Test the setup method with float value raises an exception with wrong parameters"""
        with pytest.raises(ParameterNotFound):
            yokogawa_gs200.get_parameter(parameter)

    @pytest.mark.parametrize(
        "parameter, value",
        [
            (Parameter.CURRENT, -0.001),
            (Parameter.CURRENT, 0.0005),
            (Parameter.VOLTAGE, 0.001),
            (Parameter.VOLTAGE, 0.5),
            (Parameter.RAMPING_ENABLED, True),
            (Parameter.RAMPING_ENABLED, False),
            (Parameter.RAMPING_RATE, 0.05),
            (Parameter.SOURCE_MODE, SourceMode.CURRENT),
            (Parameter.SOURCE_MODE, SourceMode.VOLTAGE),
            (Parameter.SPAN, "100mA"),
        ],
    )
    def test_set_parameter_method(self, parameter: Parameter, value, yokogawa_gs200: GS200):
        """Test the set_parameter method with float value"""
        yokogawa_gs200.set_parameter(parameter, value)
        if parameter == Parameter.SOURCE_MODE:
            assert yokogawa_gs200.source_mode == SourceMode(value)
        if parameter == Parameter.CURRENT:
            assert yokogawa_gs200.current == value
        if parameter == Parameter.VOLTAGE:
            assert yokogawa_gs200.voltage == value
        if parameter == Parameter.RAMPING_ENABLED:
            assert yokogawa_gs200.ramping_enabled == value
        if parameter == Parameter.RAMPING_RATE:
            assert yokogawa_gs200.ramping_rate == value
        if parameter == Parameter.SPAN:
            assert yokogawa_gs200.span == value

    @pytest.mark.parametrize("parameter, value", [(Parameter.MAX_CURRENT, 0.001), (Parameter.GAIN, 0.0005)])
    def test_set_parameter_method_raises_exception(self, parameter: Parameter, value, yokogawa_gs200: GS200):
        """Test the setup method with float value raises an exception with wrong parameters"""
        with pytest.raises(ParameterNotFound):
            yokogawa_gs200.set_parameter(parameter, value)

    def test_to_dict_method(self, yokogawa_gs200: GS200):
        """Test the dict method"""
        assert isinstance(yokogawa_gs200.to_dict(), dict)

    def test_turn_on_method(self, yokogawa_gs200: GS200):
        """Test start method"""
        yokogawa_gs200.turn_on()
        yokogawa_gs200.device.on.assert_called()

    def test_turn_off_method(self, yokogawa_gs200: GS200):
        """Test stop method"""
        yokogawa_gs200.turn_off()
        yokogawa_gs200.device.off.assert_called()

    @pytest.mark.parametrize("yokogawa_fixture", ["yokogawa_gs200", "yokogawa_gs200_voltage"])
    def test_initial_setup_method(self, yokogawa_fixture: str, request):
        """Test the initial setup method"""
        yokogawa_gs200 = request.getfixturevalue(yokogawa_fixture)
        yokogawa_gs200.initial_setup()
        yokogawa_gs200.device.source_mode.assert_called()
        assert yokogawa_gs200.source_mode == yokogawa_gs200.settings.source_mode
        assert yokogawa_gs200.span == yokogawa_gs200.settings.span[0]
        assert yokogawa_gs200.ramping_rate == yokogawa_gs200.settings.ramp_rate[0]
        assert yokogawa_gs200.ramping_enabled == yokogawa_gs200.settings.ramping_enabled[0]
        if yokogawa_gs200.source_mode == SourceMode.CURRENT:
            yokogawa_gs200.device.current_range.assert_called()
            if yokogawa_gs200.ramping_enabled:
                yokogawa_gs200.device.ramp_current.assert_called()
            else:
                yokogawa_gs200.device.current.assert_called()
            assert yokogawa_gs200.current == yokogawa_gs200.settings.current[0]
        else:
            yokogawa_gs200.device.voltage_range.assert_called()
            if yokogawa_gs200.ramping_enabled:
                yokogawa_gs200.device.ramp_voltage.assert_called()
            else:
                yokogawa_gs200.device.voltage.assert_called()
            assert yokogawa_gs200.voltage == yokogawa_gs200.settings.voltage[0]

    def test_ramping_enabled_property(self, yokogawa_gs200: GS200):
        """Test the source mode property"""
        assert hasattr(yokogawa_gs200, "ramping_enabled")
        assert yokogawa_gs200.ramping_enabled == yokogawa_gs200.settings.ramping_enabled[0]
        yokogawa_gs200.ramping_enabled = False
        assert not yokogawa_gs200.ramping_enabled

    def test_ramping_rate_property(self, yokogawa_gs200: GS200):
        """Test the source mode property"""
        assert hasattr(yokogawa_gs200, "ramping_rate")
        assert yokogawa_gs200.ramping_rate == yokogawa_gs200.settings.ramp_rate[0]
        yokogawa_gs200.ramping_rate = 0.05
        assert yokogawa_gs200.ramping_rate == 0.05

    def test_current_property(self, yokogawa_gs200: GS200):
        """Test the source mode property"""
        assert hasattr(yokogawa_gs200, "current")
        assert yokogawa_gs200.current == yokogawa_gs200.settings.current[0]
        yokogawa_gs200.current = 0.01
        yokogawa_gs200.device.ramp_current.assert_called()
        assert yokogawa_gs200.current == 0.01
        yokogawa_gs200.ramping_enabled = False
        yokogawa_gs200.current = 0.0
        yokogawa_gs200.device.current.assert_called()
        assert yokogawa_gs200.current == 0.0

    def test_voltage_property(self, yokogawa_gs200: GS200):
        """Test the voltage property"""
        assert hasattr(yokogawa_gs200, "voltage")
        assert yokogawa_gs200.voltage == yokogawa_gs200.settings.voltage[0]
        yokogawa_gs200.voltage = 0.01
        yokogawa_gs200.device.ramp_voltage.assert_called()
        assert yokogawa_gs200.voltage == 0.01
        yokogawa_gs200.ramping_enabled = False
        yokogawa_gs200.voltage = 0.0
        yokogawa_gs200.device.voltage.assert_called()
        assert yokogawa_gs200.voltage == 0.0

    def test_source_mode_property(self, yokogawa_gs200: GS200):
        """Test the source mode property"""
        assert hasattr(yokogawa_gs200, "source_mode")
        yokogawa_gs200.source_mode = SourceMode.VOLTAGE
        yokogawa_gs200.device.source_mode.assert_called()
        assert yokogawa_gs200.source_mode == SourceMode.VOLTAGE
        yokogawa_gs200.source_mode = SourceMode.CURRENT
        yokogawa_gs200.device.source_mode.assert_called()
        assert yokogawa_gs200.source_mode == SourceMode.CURRENT

    def test_span_property(self, yokogawa_gs200: GS200):
        """Test the span property"""
        assert hasattr(yokogawa_gs200, "span")
        assert yokogawa_gs200.span == yokogawa_gs200.settings.span[0]
        yokogawa_gs200.span = "1mA"
        yokogawa_gs200.device.current_range.assert_called()
        yokogawa_gs200.source_mode = SourceMode.VOLTAGE
        yokogawa_gs200.span = "1V"
        yokogawa_gs200.device.voltage_range.assert_called()
        assert yokogawa_gs200.span == "1V"



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Acquisitions Result"""

from dataclasses import dataclass, field

import pandas as pd

from qililab.constants import RESULTSDATAFRAME
from qililab.result.acquisition import Acquisition
from qililab.utils.dataframe_manipulation import concatenate_creating_new_name_index


@dataclass
class Acquisitions:
    """Acquisitions Results
    Args:
        acquisitions (list[Acquisition]): list of all the acquisition results

    """

    _acquisitions: list[Acquisition] = field(init=False)
    data_dataframe_indices: set[str] = field(init=False, default_factory=set)

    def acquisitions(self) -> pd.DataFrame:
        """return the acquisitions with a structure
        I, Q, Amplitude, Phase
        """
        acquisitions: list = [acquisition.acquisition for acquisition in self._acquisitions]

        return concatenate_creating_new_name_index(
            dataframes=acquisitions, new_index_name=RESULTSDATAFRAME.ACQUISITION_INDEX
        )

    def probabilities(self) -> dict[str, float]:
        """Return probabilities of being in the ground and excited state.

        Returns:
            tuple[float, float]: Probabilities of being in the ground and excited state.
        """
        raise NotImplementedError



import pytest

from qililab.analog import FluqeParameter


@pytest.fixture(name="test_parameter")
def test_parameter():
    """Dummy parameter for testing"""
    return FluqeParameter(name="foo", value=2, set_method=lambda x: 2 * x)


class TestParameter:
    """Unit tests for the ``Parameter`` class."""

    def test_init(self, test_parameter):
        """Test init method"""
        assert test_parameter.name == "foo"
        assert test_parameter._value == 2
        # test set method
        test_parameter(2)
        assert test_parameter._value == 4

    def test_get(self, test_parameter):
        """Test get method"""
        assert test_parameter() == 2

    def test_set_raw(self, test_parameter):
        """Test set_raw method"""
        test_parameter.set_raw(7)
        assert test_parameter._value == 7



from collections import deque

import numpy as np
import pytest

from qililab import Domain
from qililab.exceptions import VariableAllocated
from qililab.qprogram.blocks import Block, ForLoop, InfiniteLoop, Loop, Parallel
from qililab.qprogram.structured_program import StructuredProgram
from qililab.qprogram.variable import FloatVariable, IntVariable


class TestStructuredProgram:
    """Unit tests checking the QProgram attributes and methods"""

    @pytest.fixture
    def instance(self) -> StructuredProgram:
        return StructuredProgram()

    def test_init(self, instance: StructuredProgram):
        """Test init method"""
        assert isinstance(instance._body, Block)
        assert len(instance._body.elements) == 0
        assert isinstance(instance._block_stack, deque)
        assert len(instance._block_stack) == 1
        assert isinstance(instance._variables, dict)
        assert len(instance._variables) == 0

    def test_active_block_property(self, instance: StructuredProgram):
        """Test _active_block property"""
        assert isinstance(instance._active_block, Block)
        assert instance._active_block is instance._body

    def test_block_method(self, instance: StructuredProgram):
        """Test block method"""
        with instance.block() as block:
            # __enter__
            assert isinstance(block, Block)
            assert instance._active_block is block
        # __exit__
        assert len(instance._body.elements) == 1
        assert instance._body.elements[0] is block

    def test_infinite_loop_method(self, instance: StructuredProgram):
        """Test infinite_loop method"""
        with instance.infinite_loop() as loop:
            # __enter__
            assert isinstance(loop, InfiniteLoop)
            assert instance._active_block is loop
        # __exit__
        assert len(instance._body.elements) == 1
        assert instance._body.elements[0] is loop
        assert instance._active_block is instance._body

    def test_parallel_method(self, instance: StructuredProgram):
        """Test parallel method"""
        var1 = instance.variable(label="int_scalar", domain=Domain.Scalar, type=int)
        var2 = instance.variable(label="float_scalar", domain=Domain.Scalar, type=float)
        with instance.parallel(
            loops=[
                ForLoop(variable=var1, start=0, stop=10, step=1),
                ForLoop(variable=var2, start=0.0, stop=1.0, step=0.1),
            ]
        ) as loop:
            # __enter__
            assert isinstance(loop, Parallel)
            assert len(loop.loops) == 2
            assert instance._active_block is loop
            assert all(instance._variables[loop.variable].is_allocated for loop in loop.loops)
        # __exit__
        assert len(instance._body.elements) == 1
        assert instance._body.elements[0] is loop
        assert instance._active_block is instance._body
        assert all(not instance._variables[loop.variable].is_allocated for loop in loop.loops)

    def test_for_loop_method(self, instance: StructuredProgram):
        """Test loop method"""
        variable = instance.variable(label="int_scalar", domain=Domain.Scalar, type=int)
        start, stop, step = 0, 100, 5
        with instance.for_loop(variable=variable, start=start, stop=stop, step=step) as loop:
            # __enter__
            assert isinstance(loop, ForLoop)
            assert loop.variable == variable
            assert loop.start == start
            assert loop.stop == stop
            assert loop.step == step
            assert instance._active_block is loop
            assert instance._variables[loop.variable].is_allocated
        # __exit__
        assert len(instance._body.elements) == 1
        assert instance._body.elements[0] is loop
        assert instance._active_block is instance._body
        assert not instance._variables[loop.variable].is_allocated

    def test_loop_method(self, instance: StructuredProgram):
        """Test loop method"""
        variable = instance.variable(label="int_scalar", domain=Domain.Scalar, type=int)
        values = np.ones(10, dtype=int)
        with instance.loop(variable=variable, values=values) as loop:
            # __enter__
            assert isinstance(loop, Loop)
            assert loop.variable == variable
            assert np.array_equal(loop.values, values)
            assert instance._active_block is loop
            assert instance._variables[loop.variable].is_allocated
        # __exit__
        assert len(instance._body.elements) == 1
        assert instance._body.elements[0] is loop
        assert instance._active_block is instance._body
        assert not instance._variables[loop.variable].is_allocated

    def test_loops_raise_error_if_variable_is_allocated(self, instance: StructuredProgram):
        """Test loop method"""
        variable = instance.variable(label="int_scalar", domain=Domain.Scalar, type=int)
        # test when ForLoop allocates variable
        with instance.for_loop(variable=variable, start=0, stop=10, step=1):
            with pytest.raises(VariableAllocated):
                with instance.for_loop(variable=variable, start=100, stop=110, step=1):
                    pass
            with pytest.raises(VariableAllocated):
                with instance.loop(variable=variable, values=np.arange(10)):
                    pass
            with pytest.raises(VariableAllocated):
                with instance.parallel(loops=[Loop(variable=variable, values=np.arange(10))]):
                    pass
        # test when Loop allocates variable
        with instance.loop(variable=variable, values=np.arange(10)):
            with pytest.raises(VariableAllocated):
                with instance.for_loop(variable=variable, start=100, stop=110, step=1):
                    pass
            with pytest.raises(VariableAllocated):
                with instance.loop(variable=variable, values=np.arange(10)):
                    pass
            with pytest.raises(VariableAllocated):
                with instance.parallel(loops=[Loop(variable=variable, values=np.arange(10))]):
                    pass
        # test when Parallel allocates variable
        with instance.parallel(loops=[Loop(variable=variable, values=np.arange(10))]):
            with pytest.raises(VariableAllocated):
                with instance.for_loop(variable=variable, start=100, stop=110, step=1):
                    pass
            with pytest.raises(VariableAllocated):
                with instance.loop(variable=variable, values=np.arange(10)):
                    pass
            with pytest.raises(VariableAllocated):
                with instance.parallel(loops=[Loop(variable=variable, values=np.arange(10))]):
                    pass

    def test_variable_method(self, instance: StructuredProgram):
        """Test variable method"""
        frequency_variable = instance.variable(label="frequency", domain=Domain.Frequency)
        phase_variable = instance.variable(label="phase", domain=Domain.Phase)
        voltage_variable = instance.variable(label="voltage", domain=Domain.Voltage)
        time_variable = instance.variable(label="time", domain=Domain.Time)
        int_scalar_variable = instance.variable(label="int_scalar", domain=Domain.Scalar, type=int)
        float_scalar_variable = instance.variable(label="float_scalar", domain=Domain.Scalar, type=float)

        # Test instantiation
        assert isinstance(frequency_variable, float)
        assert isinstance(frequency_variable, FloatVariable)
        assert frequency_variable.domain is Domain.Frequency
        assert frequency_variable.label == "frequency"

        assert isinstance(phase_variable, float)
        assert isinstance(phase_variable, FloatVariable)
        assert phase_variable.domain is Domain.Phase
        assert phase_variable.label == "phase"

        assert isinstance(voltage_variable, float)
        assert isinstance(voltage_variable, FloatVariable)
        assert voltage_variable.domain is Domain.Voltage
        assert voltage_variable.label == "voltage"

        assert isinstance(time_variable, int)
        assert isinstance(time_variable, IntVariable)
        assert time_variable.domain is Domain.Time
        assert time_variable.label == "time"

        assert isinstance(int_scalar_variable, int)
        assert isinstance(int_scalar_variable, IntVariable)
        assert int_scalar_variable.domain is Domain.Scalar
        assert int_scalar_variable.label == "int_scalar"

        assert isinstance(float_scalar_variable, float)
        assert isinstance(float_scalar_variable, FloatVariable)
        assert float_scalar_variable.domain is Domain.Scalar
        assert float_scalar_variable.label == "float_scalar"

        # Test storing in program's _variables
        assert len(instance._variables) == 6
        assert all(not variable_info.is_allocated for variable_info in instance._variables.values())
        assert all(variable_info.allocated_by is None for variable_info in instance._variables.values())

    def test_variable_method_raises_error_if_domain_is_scalar_and_type_is_none(self, instance: StructuredProgram):
        """Test variable method"""
        with pytest.raises(ValueError, match="You must specify a type in a scalar variable."):
            instance.variable(label="error", domain=Domain.Scalar)

    def test_variable_method_raises_error_if_domain_is_not_scalar_and_type_is_set(self, instance: StructuredProgram):
        """Test variable method"""
        with pytest.raises(
            ValueError, match="When declaring a variable of a specific domain, its type is inferred by its domain."
        ):
            instance.variable(label="error", domain=Domain.Frequency, type=int)



"""This file tests the the ``qm_manager`` class"""

import re
from unittest.mock import MagicMock, call, patch

import numpy as np
import pytest
from qm import Program
from qm.qua import play, program

from qililab.instruments.instrument import ParameterNotFound
from qililab.instruments.quantum_machines import QuantumMachinesCluster
from qililab.settings import Settings
from qililab.typings import Parameter


@pytest.fixture(name="qua_program")
def fixture_qua_program():
    """Dummy QUA Program"""
    with program() as dummy_qua_program:
        play("pulse1", "element1")

    return dummy_qua_program


@pytest.fixture(name="qmm")
def fixture_qmm():
    """Fixture that returns an instance a qililab wrapper for Quantum Machines Manager."""
    settings = {
        "alias": "qmm",
        "address": "192.168.0.1",
        "cluster": "cluster_0",
        "controllers": [
            {
                "name": "con1",
                "analog_outputs": [
                    {"port": 1, "filter": {"feedforward": [0, 0, 0], "feedback": [0, 0, 0]}},
                    {"port": 2, "shareable": True},
                    {"port": 3},
                    {"port": 4},
                    {"port": 5},
                    {"port": 6},
                    {"port": 7},
                    {"port": 8},
                    {"port": 9},
                    {"port": 10},
                ],
                "analog_inputs": [{"port": 1}, {"port": 2}],
                "digital_outputs": [{"port": 1}, {"port": 2}, {"port": 3}, {"port": 4}, {"port": 5}],
            }
        ],
        "octaves": [],
        "elements": [
            {
                "identifier": "drive_q0",
                "mix_inputs": {
                    "I": {"controller": "con1", "port": 1},
                    "Q": {"controller": "con1", "port": 2},
                    "lo_frequency": 6e9,
                    "mixer_correction": [1.0, 0.0, 0.0, 1.0],
                },
                "intermediate_frequency": 6e9,
            },
            {
                "identifier": "readout_q0",
                "mix_inputs": {
                    "I": {"controller": "con1", "port": 3},
                    "Q": {"controller": "con1", "port": 4},
                    "lo_frequency": 6e9,
                    "mixer_correction": [1.0, 0.0, 0.0, 1.0],
                },
                "outputs": {"out1": {"controller": "con1", "port": 1}, "out2": {"controller": "con1", "port": 2}},
                "time_of_flight": 40,
                "smearing": 10,
                "threshold_rotation": 0.5,
                "threshold": 0.09,
                "intermediate_frequency": 6e9,
            },
            {"identifier": "flux_q0", "single_input": {"controller": "con1", "port": 5}},
        ],
        "run_octave_calibration": False,
    }
    qmm = QuantumMachinesCluster(settings=settings)
    qmm.device = MagicMock

    return qmm


@pytest.fixture(name="qmm_with_octave")
def fixture_qmm_with_octave():
    """Fixture that returns an instance a qililab wrapper for Quantum Machines Manager."""
    settings = {
        "alias": "qmm_with_octave",
        "address": "192.168.0.1",
        "cluster": "cluster_0",
        "controllers": [
            {
                "name": "con1",
                "analog_outputs": [
                    {"port": 1, "filter": {"feedforward": [0, 0, 0], "feedback": [0, 0, 0]}},
                    {"port": 2},
                    {"port": 3},
                    {"port": 4},
                    {"port": 5},
                    {"port": 6},
                    {"port": 7},
                    {"port": 8},
                    {"port": 9},
                    {"port": 10},
                ],
                "analog_inputs": [{"port": 1}, {"port": 2}],
                "digital_outputs": [{"port": 1}, {"port": 2}, {"port": 3}, {"port": 4}, {"port": 5}],
            }
        ],
        "octaves": [
            {
                "name": "octave1",
                "port": 11555,
                "connectivity": {"controller": "con1"},
                "loopbacks": {"Synth": "Synth2", "Dmd": "Dmd2LO"},
                "rf_outputs": [
                    {"port": 1, "lo_frequency": 6e9},
                    {"port": 2, "lo_frequency": 6e9},
                    {"port": 3, "lo_frequency": 6e9},
                    {"port": 4, "lo_frequency": 6e9},
                    {"port": 5, "lo_frequency": 6e9},
                ],
                "rf_inputs": [{"port": 1, "lo_frequency": 6e9}, {"port": 2, "lo_frequency": 6e9}],
            }
        ],
        "elements": [
            {
                "identifier": "drive_q0_rf",
                "rf_inputs": {"octave": "octave1", "port": 1},
                "digital_inputs": {"controller": "con1", "port": 1, "delay": 87, "buffer": 15},
                "digital_outputs": {"controller": "con1", "port": 1},
                "intermediate_frequency": 6e9,
            },
            {
                "identifier": "readout_q0_rf",
                "rf_inputs": {"octave": "octave1", "port": 2},
                "digital_inputs": {"controller": "con1", "port": 2, "delay": 87, "buffer": 15},
                "rf_outputs": {"octave": "octave1", "port": 1},
                "intermediate_frequency": 6e9,
                "time_of_flight": 40,
                "smearing": 10,
            },
        ],
        "run_octave_calibration": True,
    }
    qmm = QuantumMachinesCluster(settings=settings)
    qmm.device = MagicMock

    return qmm


@pytest.fixture(name="qmm_with_octave_custom_connectivity")
def fixture_qmm_with_octave_custom_connectivity():
    """Fixture that returns an instance a qililab wrapper for Quantum Machines Manager."""
    settings = {
        "alias": "qmm_with_octave_custom_connectivity",
        "address": "192.168.0.1",
        "cluster": "cluster_0",
        "controllers": [
            {
                "name": "con1",
                "analog_outputs": [
                    {"port": 1, "filter": {"feedforward": [0, 0, 0], "feedback": [0, 0, 0]}},
                    {"port": 2},
                    {"port": 3},
                    {"port": 4},
                    {"port": 5},
                    {"port": 6},
                    {"port": 7},
                    {"port": 8},
                    {"port": 9},
                    {"port": 10},
                ],
                "analog_inputs": [{"port": 1}, {"port": 2}],
                "digital_outputs": [{"port": 1}, {"port": 2}, {"port": 3}, {"port": 4}, {"port": 5}],
            }
        ],
        "octaves": [
            {
                "name": "octave1",
                "port": 11555,
                "rf_outputs": [
                    {
                        "port": 1,
                        "lo_frequency": 6e9,
                        "i_connection": {"controller": "con1", "port": 1},
                        "q_connection": {"controller": "con1", "port": 2},
                    },
                    {
                        "port": 2,
                        "lo_frequency": 6e9,
                        "i_connection": {"controller": "con1", "port": 3},
                        "q_connection": {"controller": "con1", "port": 4},
                    },
                    {
                        "port": 3,
                        "lo_frequency": 6e9,
                        "i_connection": {"controller": "con1", "port": 5},
                        "q_connection": {"controller": "con1", "port": 6},
                    },
                    {
                        "port": 4,
                        "lo_frequency": 6e9,
                        "i_connection": {"controller": "con1", "port": 7},
                        "q_connection": {"controller": "con1", "port": 8},
                    },
                    {
                        "port": 5,
                        "lo_frequency": 6e9,
                        "i_connection": {"controller": "con1", "port": 9},
                        "q_connection": {"controller": "con1", "port": 10},
                    },
                ],
                "rf_inputs": [{"port": 1, "lo_frequency": 6e9}, {"port": 2, "lo_frequency": 6e9}],
                "if_outputs": [{"controller": "con1", "port": 1}, {"controller": "con1", "port": 2}],
                "loopbacks": {"Synth": "Synth2", "Dmd": "Dmd2LO"},
            }
        ],
        "elements": [
            {
                "identifier": "drive_q0_rf",
                "rf_inputs": {"octave": "octave1", "port": 1},
                "digital_inputs": {"controller": "con1", "port": 1, "delay": 87, "buffer": 15},
                "digital_outputs": {"controller": "con1", "port": 1},
                "intermediate_frequency": 6e9,
            },
            {
                "identifier": "readout_q0_rf",
                "rf_inputs": {"octave": "octave1", "port": 2},
                "digital_inputs": {"controller": "con1", "port": 2, "delay": 87, "buffer": 15},
                "rf_outputs": {"octave": "octave1", "port": 1},
                "intermediate_frequency": 6e9,
                "time_of_flight": 40,
                "smearing": 10,
            },
        ],
        "run_octave_calibration": True,
    }
    qmm = QuantumMachinesCluster(settings=settings)
    qmm.device = MagicMock

    return qmm


@pytest.fixture(name="qmm_with_opx1000")
def fixture_qmm_with_opx1000():
    """Fixture that returns an instance a qililab wrapper for Quantum Machines Manager."""
    settings = {
        "alias": "qmm_with_opx1000",
        "address": "192.168.0.1",
        "cluster": "cluster_0",
        "controllers": [
            {
                "name": "con1",
                "type": "opx1000",
                "fems": [
                    {
                        "fem": 1,
                        "analog_outputs": [
                            {"port": 1, "filter": {"feedforward": [0, 0, 0], "feedback": [0, 0, 0]}},
                            {"port": 2},
                            {"port": 3},
                            {"port": 4},
                            {"port": 5},
                            {"port": 6},
                            {"port": 7},
                            {"port": 8},
                        ],
                        "analog_inputs": [{"port": 1}, {"port": 2}],
                        "digital_outputs": [{"port": 1}, {"port": 2}, {"port": 3}, {"port": 4}, {"port": 5}],
                    }
                ],
            }
        ],
        "octaves": [
            {
                "name": "octave1",
                "port": 11555,
                "connectivity": {"controller": "con1", "fem": 1},
                "loopbacks": {"Synth": "Synth2", "Dmd": "Dmd2LO"},
                "rf_outputs": [
                    {
                        "port": 1,
                        "lo_frequency": 6e9,
                        "i_connection": {"controller": "con1", "fem": 1, "port": 1},
                        "q_connection": {"controller": "con1", "fem": 1, "port": 2},
                    },
                    {
                        "port": 2,
                        "lo_frequency": 6e9,
                        "i_connection": {"controller": "con1", "fem": 1, "port": 3},
                        "q_connection": {"controller": "con1", "fem": 1, "port": 4},
                    },
                    {
                        "port": 3,
                        "lo_frequency": 6e9,
                        "i_connection": {"controller": "con1", "fem": 1, "port": 5},
                        "q_connection": {"controller": "con1", "fem": 1, "port": 6},
                    },
                    {
                        "port": 4,
                        "lo_frequency": 6e9,
                        "i_connection": {"controller": "con1", "fem": 1, "port": 7},
                        "q_connection": {"controller": "con1", "fem": 1, "port": 8},
                    },
                    {
                        "port": 5,
                        "lo_frequency": 6e9,
                        "i_connection": {"controller": "con1", "fem": 1, "port": 9},
                        "q_connection": {"controller": "con1", "fem": 1, "port": 10},
                    },
                ],
                "rf_inputs": [{"port": 1, "lo_frequency": 6e9}, {"port": 2, "lo_frequency": 6e9}],
            }
        ],
        "elements": [
            {
                "identifier": "drive_q0_rf",
                "rf_inputs": {"octave": "octave1", "port": 1},
                "digital_inputs": {"controller": "con1", "port": 1, "delay": 87, "buffer": 15},
                "digital_outputs": {"controller": "con1", "fem": 1, "port": 1},
                "intermediate_frequency": 6e9,
            },
            {
                "identifier": "readout_q0_rf",
                "rf_inputs": {"octave": "octave1", "port": 2},
                "digital_inputs": {"controller": "con1", "port": 2, "delay": 87, "buffer": 15},
                "rf_outputs": {"octave": "octave1", "port": 1},
                "intermediate_frequency": 6e9,
                "time_of_flight": 40,
                "smearing": 10,
            },
            {"identifier": "flux_q0", "single_input": {"controller": "con1", "fem": 1, "port": 5}},
        ],
        "run_octave_calibration": True,
    }
    qmm = QuantumMachinesCluster(settings=settings)
    qmm.device = MagicMock

    return qmm


@pytest.fixture(name="qmm_int_identifiers")
def fixture_qmm_int_identifiers():
    """Fixture that returns an instance a qililab wrapper for Quantum Machines Manager."""
    settings = {
        "alias": "qmm",
        "address": "192.168.0.1",
        "cluster": "cluster_0",
        "controllers": [
            {
                "name": "con1",
                "analog_outputs": [
                    {"port": 1, "filter": {"feedforward": [0, 0, 0], "feedback": [0, 0, 0]}},
                    {"port": 2},
                    {"port": 3},
                    {"port": 4},
                    {"port": 5},
                    {"port": 6},
                    {"port": 7},
                    {"port": 8},
                    {"port": 9},
                    {"port": 10},
                ],
                "analog_inputs": [{"port": 1}, {"port": 2}],
                "digital_outputs": [{"port": 1}, {"port": 2}, {"port": 3}, {"port": 4}, {"port": 5}],
            }
        ],
        "octaves": [],
        "elements": [
            {
                "identifier": 1,
                "mix_inputs": {
                    "I": {"controller": "con1", "port": 1},
                    "Q": {"controller": "con1", "port": 2},
                    "lo_frequency": 6e9,
                    "mixer_correction": [1.0, 0.0, 0.0, 1.0],
                },
                "intermediate_frequency": 6e9,
            },
            {
                "identifier": 0,
                "mix_inputs": {
                    "I": {"controller": "con1", "port": 3},
                    "Q": {"controller": "con1", "port": 4},
                    "lo_frequency": 6e9,
                    "mixer_correction": [1.0, 0.0, 0.0, 1.0],
                },
                "outputs": {"out1": {"controller": "con1", "port": 1}, "out2": {"controller": "con1", "port": 2}},
                "time_of_flight": 40,
                "smearing": 10,
                "threshold_rotation": 0.5,
                "threshold": 0.09,
                "intermediate_frequency": 6e9,
            },
            {"identifier": "flux_q0", "single_input": {"controller": "con1", "port": 5}},
        ],
        "run_octave_calibration": False,
    }
    qmm = QuantumMachinesCluster(settings=settings)
    qmm.device = MagicMock

    return qmm


@pytest.fixture(name="compilation_config")
def fixture_compilation_config() -> dict:
    """Fixture that returns a configuration dictionary as the QuantumMachinesCompiler would."""
    config = {
        "elements": {
            "drive_q0": {
                "operations": {"control_445e964c_fb58e912_100": "control_445e964c_fb58e912_100"},
                "RF_inputs": {"port": ("octave1", 1)},
            }
        },
        "pulses": {
            "control_445e964c_fb58e912_100": {
                "operation": "control",
                "length": 100,
                "waveforms": {"I": "445e964c", "Q": "fb58e912"},
            },
        },
        "waveforms": {"445e964c": {"type": "constant", "sample": 1.0}, "fb58e912": {"type": "constant", "sample": 0.0}},
        "octaves": {"octave1": {"RF_outputs": {1: {"gain": 0.5}}}},
    }
    return config


class MockJob:
    """Mocks a running job from Quantum Machines"""

    def __init__(self):
        self.result_handles = MockStreamingFetcher()


class MockStreamingFetcher:
    """Mocks the StreamingFetcher class from Quantum Machines."""

    def __init__(self):
        self.values = [("I", MockSingleHandle()), ("Q", MockSingleHandle())]
        self.index = 0

    def wait_for_all_values(self, timeout: int | None = None):
        """Mocks waiting for all values method from streamer"""
        return MagicMock

    def __iter__(self):
        return self

    def __next__(self):
        if self.index < len(self.values):
            result = self.values[self.index]
            self.index += 1
            return result

        raise StopIteration


class MockSingleHandle:
    """Mocks a single result handle from quantum machines results module."""

    def __init__(self):
        self.values = np.zeros((10))
        self.index = 0

    def fetch_all(self, flat_struct: bool):
        """Mocks fetching all values from the result handle."""
        return self.values


class TestQuantumMachinesCluster:
    """This class contains the unit tests for the ``QuantumMachinesCluster`` class."""

    @patch("qililab.instruments.quantum_machines.quantum_machines_cluster.QuantumMachinesManager")
    @patch("qililab.instruments.Instrument.initial_setup")
    @pytest.mark.parametrize(
        "qmm_name",
        ["qmm", "qmm_with_octave", "qmm_with_octave_custom_connectivity", "qmm_with_opx1000", "qmm_int_identifiers"],
    )
    def test_initial_setup(self, mock_instrument_init: MagicMock, mock_init: MagicMock, qmm_name, request):
        """Test QMM class initialization."""
        qmm = request.getfixturevalue(qmm_name)

        # Before initial_setup no _config should exist
        assert qmm._config_created is False and "_config" not in dir(qmm)

        qmm.initial_setup()
        mock_init.assert_called()

        assert isinstance(qmm._qmm, MagicMock)
        assert isinstance(qmm._config, dict)
        assert isinstance(qmm.config, dict)

        # Assert that the _config has been created correctly (in synch):
        assert qmm._config_created is True
        assert qmm._config == qmm.settings.to_qua_config()

    @pytest.mark.parametrize(
        "qmm_name",
        ["qmm", "qmm_with_octave", "qmm_with_octave_custom_connectivity", "qmm_with_opx1000", "qmm_int_identifiers"],
    )
    def test_settings(self, qmm_name, request):
        """Test QuantumMachinesClusterSettings have been set correctly"""

        qmm = request.getfixturevalue(qmm_name)
        assert isinstance(qmm.settings, Settings)
        assert qmm.is_awg()
        assert qmm.is_adc()

    @patch("qililab.instruments.quantum_machines.quantum_machines_cluster.QuantumMachinesManager")
    @patch("qililab.instruments.quantum_machines.quantum_machines_cluster.QuantumMachine")
    @pytest.mark.parametrize(
        "qmm_name",
        ["qmm", "qmm_with_octave", "qmm_with_octave_custom_connectivity", "qmm_with_opx1000", "qmm_int_identifiers"],
    )
    def test_turn_on(self, mock_qmm, mock_qm, qmm_name, request):
        """Test turn_on method"""

        qmm = request.getfixturevalue(qmm_name)
        qmm.initial_setup()
        qmm.turn_on()

        assert isinstance(qmm._qm, MagicMock)
        if qmm.settings.run_octave_calibration:
            calls = [
                call(element) for element in qmm._config["elements"] if "RF_inputs" in qmm._config["elements"][element]
            ]
            qmm._qm.calibrate_element.assert_has_calls(calls)

        # Assert that the settings are still in synch:
        assert qmm._config == qmm.settings.to_qua_config()

    @patch("qililab.instruments.quantum_machines.quantum_machines_cluster.QuantumMachinesManager")
    @patch("qililab.instruments.quantum_machines.quantum_machines_cluster.QuantumMachine")
    @pytest.mark.parametrize(
        "qmm_name",
        ["qmm", "qmm_with_octave", "qmm_with_octave_custom_connectivity", "qmm_with_opx1000", "qmm_int_identifiers"],
    )
    def test_turn_off(self, mock_qmm, mock_qm, qmm_name, request):
        """Test turn_off method"""

        qmm = request.getfixturevalue(qmm_name)
        qmm.initial_setup()
        qmm.turn_on()
        qmm.turn_off()

        # Assert that the settings are still in synch:
        assert qmm._config == qmm.settings.to_qua_config()

        assert isinstance(qmm._qm, MagicMock)
        qmm._qm.close.assert_called_once()

    @patch("qililab.instruments.quantum_machines.quantum_machines_cluster.QuantumMachinesManager")
    @patch("qililab.instruments.quantum_machines.quantum_machines_cluster.QuantumMachine")
    def test_append_configuration(self, mock_qmm, mock_qm, qmm: QuantumMachinesCluster, compilation_config: dict):
        """Test update_configuration method"""
        qmm.initial_setup()
        qmm.append_configuration(configuration=compilation_config)

        assert "control_445e964c_fb58e912_100" in qmm._config["elements"]["drive_q0"]["operations"]
        assert "control_445e964c_fb58e912_100" in qmm._config["pulses"]
        assert "445e964c" in qmm._config["waveforms"]
        assert "fb58e912" in qmm._config["waveforms"]

        # Assert that the `settings.to_qua_config()`` are a subset of the `appended_configuration``, and in synch:
        for k, v in qmm.settings.to_qua_config().items():
            if k != "elements" and v:
                assert v == qmm._config[k]
            if k == "elements":
                for element in v:
                    for key, value in v[element].items():
                        if value:
                            assert value == qmm._config[k][element][key]

    @patch("qililab.instruments.quantum_machines.quantum_machines_cluster.QuantumMachinesManager")
    @patch("qililab.instruments.quantum_machines.quantum_machines_cluster.QuantumMachine")
    def test_append_configuration_after_turn_on(
        self, mock_qmm, mock_qm, qmm: QuantumMachinesCluster, compilation_config: dict
    ):
        """Test update_configuration method"""
        qmm.initial_setup()
        qmm.turn_on()
        qmm.append_configuration(configuration=compilation_config)

        qmm._qmm.open_qm.call_count == 2
        assert isinstance(qmm._qm, MagicMock)

        # Assert that the `settings.to_qua_config()`` are a subset of the `appended_configuration``, and in synch:
        for k, v in qmm.settings.to_qua_config().items():
            if k != "elements" and v:
                assert v == qmm._config[k]
            if k == "elements":
                for element in v:
                    for key, value in v[element].items():
                        if value:
                            assert value == qmm._config[k][element][key]

    @patch("qililab.instruments.quantum_machines.quantum_machines_cluster.QuantumMachinesManager")
    @patch("qililab.instruments.quantum_machines.quantum_machines_cluster.QuantumMachine")
    def test_append_configuration_without_initial_setup_raises_error(
        self, mock_qmm, mock_qm, qmm: QuantumMachinesCluster, compilation_config: dict
    ):
        """Test update_configuration method raises an error when no config dict has been set."""
        with pytest.raises(
            ValueError,
            match=re.escape("The QM `config` dictionary does not exist. Please run `initial_setup()` first."),
        ):
            qmm.append_configuration(configuration=compilation_config)

    @patch("qililab.instruments.quantum_machines.quantum_machines_cluster.QuantumMachinesManager")
    @patch("qililab.instruments.quantum_machines.quantum_machines_cluster.QuantumMachine")
    def test_get_controller_type_from_bus_singleInput(
        self, mock_qmm, mock_qm, qmm: QuantumMachinesCluster, compilation_config: dict
    ):
        qmm.initial_setup()
        qmm.turn_on()

        controller = qmm.get_controller_type_from_bus("flux_q0")
        assert controller == "opx1"

    @patch("qililab.instruments.quantum_machines.quantum_machines_cluster.QuantumMachinesManager")
    @patch("qililab.instruments.quantum_machines.quantum_machines_cluster.QuantumMachine")
    def test_get_controller_type_from_bus_without_controller_raises_error(
        self, mock_qmm, mock_qm, qmm: QuantumMachinesCluster, compilation_config: dict
    ):
        """Test get_controller_type_from_bus method raises an error when no controller is inside bus."""
        qmm.initial_setup()
        qmm.turn_on()

        qmm._config["elements"]["bus"] = {"singleInput": {"port": ("con10", 1)}}

        with pytest.raises(
            AttributeError,
            match=re.escape("Controller with bus bus does not exist"),
        ):
            qmm.get_controller_type_from_bus("bus")

    @patch("qililab.instruments.quantum_machines.quantum_machines_cluster.QuantumMachinesManager")
    @patch("qililab.instruments.quantum_machines.quantum_machines_cluster.QuantumMachine")
    def test_get_controller_from_element_wrong_key_raises_error(
        self, mock_qmm, mock_qm, qmm: QuantumMachinesCluster, compilation_config: dict
    ):
        """Test get_controller_type_from_bus method raises an error when no controller is inside bus."""
        qmm.initial_setup()
        qmm.turn_on()

        element = next((element for element in qmm.settings.elements if element["identifier"] == "readout_q0"), None)

        with pytest.raises(
            ValueError,
            match=re.escape("key value must be I or Q, O given"),
        ):
            qmm.get_controller_from_element(element=element, key="O")

    @patch("qililab.instruments.quantum_machines.quantum_machines_cluster.QuantumMachinesManager")
    @patch("qililab.instruments.quantum_machines.quantum_machines_cluster.QuantumMachine")
    def test_compile(self, mock_qmm, mock_qm, qmm: QuantumMachinesCluster, qua_program: Program):
        qmm.initial_setup()
        qmm.turn_on()

        assert len(qmm._compiled_program_cache) == 0
        qmm._qm.compile.return_value = "123"

        compile_program_id = qmm.compile(qua_program)
        assert len(qmm._compiled_program_cache) == 1
        assert compile_program_id == "123"

        compile_program_id = qmm.compile(qua_program)
        assert len(qmm._compiled_program_cache) == 1
        assert compile_program_id == "123"

        # Assert that the settings are still in synch:
        assert qmm._config == qmm.settings.to_qua_config()

    @patch("qm.QuantumMachine")
    def test_run(self, mock_qm: MagicMock, qmm: QuantumMachinesCluster, qua_program: Program):
        """Test execute method"""
        mock_qm.return_value.execute.return_value = MagicMock
        qmm._qm = mock_qm
        job = qmm.run(qua_program)

        assert isinstance(job, MagicMock)

        # Assert that the settings are in synch:
        assert qmm._config_created is False and "_config" not in dir(qmm)

    @patch("qililab.instruments.quantum_machines.quantum_machines_cluster.QuantumMachinesManager")
    @patch("qililab.instruments.quantum_machines.quantum_machines_cluster.QuantumMachine")
    def test_run_compiled_program(self, mock_qmm, mock_qm, qmm: QuantumMachinesCluster, qua_program: Program):
        qmm.initial_setup()
        qmm.turn_on()

        qmm._qm.compile.return_value = "123"
        qmm._controller = "opx1000"  # type: ignore[attr-defined]
        qmm._pending_set_intermediate_frequency = {"drive_q0": 20e6}

        compile_program_id = qmm.compile(qua_program)
        _ = qmm.run_compiled_program(compile_program_id)

        # TODO: qm.queue.add_compiled() -> qm.add_compiled()
        qmm._qm.queue.add_compiled.assert_called_once_with(compile_program_id)
        # TODO: job.wait_for_execution() is deprecated and will be removed in the future. Please use job.wait_until("Running") instead.
        # The following stopped working in testing, but we have verified that works in hardware, so I remove it temporarily.
        # qmm._qm.queue.add_compiled.return_value.wait_until.assert_called_once()

        qmm._qm.calibrate_element.assert_called_once()
        # Assert that the settings are still in synch:
        assert qmm._config == qmm.settings.to_qua_config()

    def test_get_acquisitions(self, qmm: QuantumMachinesCluster):
        """Test get_acquisitions method"""
        job = MockJob()
        results = qmm.get_acquisitions(job)

        assert isinstance(results, dict)
        assert "I" in results
        assert "Q" in results

        # Assert that the settings are in synch:
        assert qmm._config_created is False and "_config" not in dir(qmm)

    @patch("qm.QuantumMachine")
    def test_simulate(self, mock_qm: MagicMock, qmm: QuantumMachinesCluster, qua_program: Program):
        """Test simulate method"""
        mock_qm.return_value.simulate.return_value = MagicMock
        qmm._qm = mock_qm
        job = qmm.simulate(qua_program)

        assert isinstance(job, MagicMock)

        # Assert that the settings are in synch:
        assert qmm._config_created is False and "_config" not in dir(qmm)

    @pytest.mark.parametrize(
        "bus, parameter, value",
        [
            ("drive_q0_rf", Parameter.LO_FREQUENCY, 6e9),
            ("drive_q0_rf", Parameter.IF, 20e6),
            ("drive_q0_rf", Parameter.GAIN, 0.001),
            ("readout_q0_rf", Parameter.LO_FREQUENCY, 6e9),
            ("readout_q0_rf", Parameter.IF, 20e6),
            ("readout_q0_rf", Parameter.GAIN, 0.001),
            ("readout_q0_rf", Parameter.OFFSET_I, 0.001),
            ("drive_q0_rf", Parameter.OFFSET_Q, 0.001),
            ("readout_q0_rf", Parameter.OFFSET_OUT1, 0.001),
            ("readout_q0_rf", Parameter.OFFSET_OUT2, 0.001),
        ],
    )
    @patch("qililab.instruments.quantum_machines.quantum_machines_cluster.QuantumMachinesManager")
    @patch("qililab.instruments.quantum_machines.quantum_machines_cluster.QuantumMachine")
    def test_set_parameter_method_with_octave(
        self,
        mock_qmm,
        mock_qm,
        bus: str,
        parameter: Parameter,
        value: float | str | bool,
        qmm_with_octave: QuantumMachinesCluster,
    ):
        """Test the setup method with float value"""
        qmm_with_octave.initial_setup()
        qmm_with_octave.turn_on()
        qmm_with_octave._config = qmm_with_octave.settings.to_qua_config()

        qmm_with_octave.set_parameter(parameter=parameter, value=value, channel_id=bus)
        if parameter == Parameter.LO_FREQUENCY:
            qmm_with_octave._qm.octave.set_lo_frequency.assert_called_once()
            calls = [
                call(element)
                for element in qmm_with_octave._config["elements"]
                if "RF_inputs" in qmm_with_octave._config["elements"][element]
            ]
            qmm_with_octave._qm.calibrate_element.assert_has_calls(calls)
        if parameter == Parameter.GAIN:
            qmm_with_octave._qm.octave.set_rf_output_gain.assert_called_once()
        if parameter == Parameter.IF:
            qmm_with_octave._qm.set_intermediate_frequency.assert_called_once()
        if parameter == Parameter.OFFSET_I:
            qmm_with_octave._qm.set_output_dc_offset_by_element.assert_called_once()
        if parameter == Parameter.OFFSET_Q:
            qmm_with_octave._qm.set_output_dc_offset_by_element.assert_called_once()
        if parameter == Parameter.OFFSET_OUT1:
            qmm_with_octave._qm.set_input_dc_offset_by_element.assert_called_once()
        if parameter == Parameter.OFFSET_OUT2:
            qmm_with_octave._qm.set_input_dc_offset_by_element.assert_called_once()

        # Assert that the settings are still in synch:
        assert qmm_with_octave._config == qmm_with_octave.settings.to_qua_config()

    @pytest.mark.parametrize(
        "bus, parameter, value",
        [
            ("drive_q0", Parameter.IF, 20e6),
            ("readout_q0", Parameter.THRESHOLD_ROTATION, 0.5),
            ("readout_q0", Parameter.THRESHOLD, 0.01),
            ("flux_q0", Parameter.DC_OFFSET, 0.001),
            ("readout_q0", Parameter.OFFSET_I, 0.001),
            ("drive_q0", Parameter.OFFSET_Q, 0.001),
            ("readout_q0", Parameter.OFFSET_OUT1, 0.001),
            ("readout_q0", Parameter.OFFSET_OUT2, 0.001),
        ],
    )
    @patch("qililab.instruments.quantum_machines.quantum_machines_cluster.QuantumMachinesManager")
    @patch("qililab.instruments.quantum_machines.quantum_machines_cluster.QuantumMachine")
    def test_set_parameter_method(
        self, mock_qmm, mock_qm, bus: str, parameter: Parameter, value: float | str | bool, qmm: QuantumMachinesCluster
    ):
        """Test the setup method with float value"""
        qmm.initial_setup()
        qmm.turn_on()
        qmm._config = qmm.settings.to_qua_config()

        qmm.set_parameter(parameter=parameter, value=value, channel_id=bus)

        element = next((element for element in qmm.settings.elements if element["identifier"] == bus))
        if parameter == Parameter.IF:
            assert value == element["intermediate_frequency"]
        if parameter == Parameter.THRESHOLD_ROTATION:
            assert value == element["threshold_rotation"]
        if parameter == Parameter.THRESHOLD:
            assert value == element["threshold"]
        if parameter == Parameter.DC_OFFSET:
            qmm._qm.set_output_dc_offset_by_element.assert_called_once()
        if parameter == Parameter.OFFSET_I:
            qmm._qm.set_output_dc_offset_by_element.assert_called_once()
        if parameter == Parameter.OFFSET_Q:
            qmm._qm.set_output_dc_offset_by_element.assert_called_once()
        if parameter == Parameter.OFFSET_OUT1:
            qmm._qm.set_input_dc_offset_by_element.assert_called_once()
        if parameter == Parameter.OFFSET_OUT2:
            qmm._qm.set_input_dc_offset_by_element.assert_called_once()

        # Assert that the settings are still in synch:
        assert qmm._config == qmm.settings.to_qua_config()

    @pytest.mark.parametrize(
        "bus, parameter, value",
        [
            ("drive_q0", Parameter.IF, 17e6),
        ],
    )
    @patch("qililab.instruments.quantum_machines.quantum_machines_cluster.QuantumMachinesManager")
    @patch("qililab.instruments.quantum_machines.quantum_machines_cluster.QuantumMachine")
    def test_set_parameter_without_connection_changes_settings(
        self, mock_qmm, mock_qm, bus: str, parameter: Parameter, value: float | str | bool, qmm: QuantumMachinesCluster
    ):
        """Test that both the local `settings` and `_config` are changed by the set method without connection."""

        # Set intermidiate frequency to 17e6 locally
        qmm.set_parameter(parameter=parameter, value=value, channel_id=bus)
        qmm.initial_setup()

        # Test that both the local `settings` and `_config` have been changed to 17e6:
        assert qmm._config == qmm.settings.to_qua_config()
        # Test `_config` QUA dictionary:
        assert qmm._config["elements"][bus][parameter] == value
        # Test `settings` qililab dictionary:
        assert qmm.settings.to_qua_config()["elements"][bus][parameter] == value

        # Assert that the settings are still in synch:
        assert qmm._config == qmm.settings.to_qua_config()

    @pytest.mark.parametrize(
        "bus, parameter, value",
        [
            ("readout_q0_rf", Parameter.IF, 17e6),
            ("flux_q0", Parameter.DC_OFFSET, 0.001),
            ("readout_q0_rf", Parameter.OFFSET_I, 0.001),
            ("drive_q0_rf", Parameter.OFFSET_Q, 0.001),
            ("readout_q0_rf", Parameter.OFFSET_OUT1, 0.001),
            ("readout_q0_rf", Parameter.OFFSET_OUT2, 0.001),
        ],
    )
    @patch("qililab.instruments.quantum_machines.quantum_machines_cluster.QuantumMachinesManager")
    @patch("qililab.instruments.quantum_machines.quantum_machines_cluster.QuantumMachine")
    def test_set_parameter_with_opx1000(
        self,
        mock_qmm,
        mock_qm,
        bus: str,
        parameter: Parameter,
        value: float | str | bool,
        qmm_with_opx1000: QuantumMachinesCluster,
    ):
        """Test that both the set method fills the bus _pending_set_intermediate_frequency correctly."""

        qmm_with_opx1000.initial_setup()
        qmm_with_opx1000.turn_on()
        qmm_with_opx1000._config = qmm_with_opx1000.settings.to_qua_config()

        qmm_with_opx1000.set_parameter(parameter=parameter, value=value, channel_id=bus)

        if parameter == Parameter.IF:
            # Test `_intermediate_frequency[bus]` is created for later use:
            assert qmm_with_opx1000._pending_set_intermediate_frequency[bus] == value
        if parameter == Parameter.DC_OFFSET:
            qmm_with_opx1000._qm.set_output_dc_offset_by_element.assert_called_once()
        if parameter == Parameter.OFFSET_I:
            qmm_with_opx1000._qm.set_output_dc_offset_by_element.assert_called_once()
        if parameter == Parameter.OFFSET_Q:
            qmm_with_opx1000._qm.set_output_dc_offset_by_element.assert_called_once()
        if parameter == Parameter.OFFSET_OUT1:
            qmm_with_opx1000._qm.set_input_dc_offset_by_element.assert_called_once()
        if parameter == Parameter.OFFSET_OUT2:
            qmm_with_opx1000._qm.set_input_dc_offset_by_element.assert_called_once()

    @pytest.mark.parametrize(
        "bus, parameter, value",
        [
            ("drive_q0", Parameter.LO_FREQUENCY, 6e9),
            ("drive_q0", Parameter.GAIN, 0.001),
        ],
    )
    @patch("qililab.instruments.quantum_machines.quantum_machines_cluster.QuantumMachinesManager")
    @patch("qililab.instruments.quantum_machines.quantum_machines_cluster.QuantumMachine")
    def test_set_parameter_method_raises_error_when_parameter_is_for_octave_and_there_is_no_octave(
        self, mock_qmm, mock_qm, bus: str, parameter: Parameter, value: float | str | bool, qmm: QuantumMachinesCluster
    ):
        """Test the set_parameter method raises exception when the parameter is for octave and there is no octave connected to the bus."""
        qmm.initial_setup()
        qmm.turn_on()
        qmm._config = qmm.settings.to_qua_config()

        with pytest.raises(
            ValueError,
            match=f"Trying to change parameter {parameter.name} in {qmm.name}, however bus {bus} is not connected to an octave.",
        ):
            qmm.set_parameter(parameter=parameter, value=value, channel_id=bus)

        # Assert that the settings are still in synch:
        assert qmm._config == qmm.settings.to_qua_config()

    @pytest.mark.parametrize(
        "parameter, value", [(Parameter.LO_FREQUENCY, 6e9), (Parameter.MAX_CURRENT, 0.5), (Parameter.OUT0_ATT, 0.01)]
    )
    @patch("qililab.instruments.quantum_machines.quantum_machines_cluster.QuantumMachinesManager")
    @patch("qililab.instruments.quantum_machines.quantum_machines_cluster.QuantumMachine")
    def test_set_parameter_method_raises_exception_when_bus_not_found(
        self, mock_qmm, mock_qm, parameter: Parameter, value: float | str | bool, qmm: QuantumMachinesCluster
    ):
        """Test the set_parameter method raises exception when parameter is wrong."""
        non_existent_bus = "non_existent_bus"

        qmm.initial_setup()
        qmm.turn_on()
        with pytest.raises(ValueError, match=f"Bus {non_existent_bus} was not found in {qmm.name} settings."):
            qmm.set_parameter(parameter, value, channel_id=non_existent_bus)

        # Assert that the settings are still in synch:
        assert qmm._config == qmm.settings.to_qua_config()

    @pytest.mark.parametrize("parameter, value", [(Parameter.MAX_CURRENT, 0.001), (Parameter.OUT0_ATT, 0.0005)])
    @patch("qililab.instruments.quantum_machines.quantum_machines_cluster.QuantumMachinesManager")
    @patch("qililab.instruments.quantum_machines.quantum_machines_cluster.QuantumMachine")
    def test_set_parameter_method_raises_exception_when_parameter_not_found(
        self, mock_qmm, mock_qm, parameter: Parameter, value, qmm: QuantumMachinesCluster
    ):
        """Test the set_parameter method raises exception when parameter is wrong."""
        qmm.initial_setup()
        qmm.turn_on()
        with pytest.raises(ParameterNotFound, match=f"Could not find parameter {parameter} in instrument {qmm.name}."):
            qmm.set_parameter(parameter, value, channel_id="drive_q0")

        # Assert that the settings are still in synch:
        assert qmm._config == qmm.settings.to_qua_config()

    @pytest.mark.parametrize(
        "bus, parameter, qmm_name",
        [
            ("drive_q0", Parameter.LO_FREQUENCY, "qmm"),
            ("drive_q0_rf", Parameter.LO_FREQUENCY, "qmm_with_octave"),
            ("drive_q0", Parameter.IF, "qmm"),
            ("readout_q0", Parameter.GAIN, "qmm"),
            ("drive_q0_rf", Parameter.GAIN, "qmm_with_octave"),
            ("readout_q0", Parameter.TIME_OF_FLIGHT, "qmm"),
            ("readout_q0", Parameter.SMEARING, "qmm"),
            ("readout_q0", Parameter.THRESHOLD_ROTATION, "qmm"),
            ("readout_q0", Parameter.THRESHOLD, "qmm"),
            ("flux_q0", Parameter.DC_OFFSET, "qmm"),
            ("readout_q0", Parameter.OFFSET_I, "qmm"),
            ("drive_q0_rf", Parameter.OFFSET_Q, "qmm_with_octave"),
            ("readout_q0", Parameter.OFFSET_OUT1, "qmm"),
            ("readout_q0", Parameter.OFFSET_OUT2, "qmm"),
        ],
    )
    @patch("qililab.instruments.quantum_machines.quantum_machines_cluster.QuantumMachinesManager")
    @patch("qililab.instruments.quantum_machines.quantum_machines_cluster.QuantumMachine")
    def test_get_parameter_method(
        self,
        mock_qmm,
        mock_qm,
        bus: str,
        parameter: Parameter,
        qmm_name: QuantumMachinesCluster,
        request,
    ):
        """Test the setup method with float value"""
        qmm = request.getfixturevalue(qmm_name)
        value = qmm.get_parameter(parameter, channel_id=bus)

        settings_config_dict = qmm.settings.to_qua_config()

        config_keys = settings_config_dict["elements"][bus]

        if parameter == Parameter.LO_FREQUENCY:
            if "mixInputs" in config_keys:
                assert value == settings_config_dict["elements"][bus]["mixInputs"]["lo_frequency"]
            if "RF_inputs" in config_keys:
                port = settings_config_dict["elements"][bus]["RF_inputs"]["port"]
                assert value == settings_config_dict["octaves"][port[0]]["RF_outputs"][port[1]]["LO_frequency"]
        if parameter == Parameter.IF:
            if "intermediate_frequency" in config_keys:
                assert value == settings_config_dict["elements"][bus]["intermediate_frequency"]
        if parameter == Parameter.GAIN:
            if "mixInputs" in config_keys and "outputs" in config_keys:
                port_i = settings_config_dict["elements"][bus]["outputs"]["out1"]
                port_q = settings_config_dict["elements"][bus]["outputs"]["out2"]
                assert value == settings_config_dict["controllers"][port_i[0]]["analog_inputs"][port_i[1]]["gain_db"]
                # assert value == (
                #     settings_config_dict["controllers"][port_i[0]]["analog_inputs"][port_i[1]]["gain_db"],
                #     settings_config_dict["controllers"][port_q[0]]["analog_inputs"][port_q[1]]["gain_db"],
                # )
            if "RF_inputs" in config_keys:
                port = settings_config_dict["elements"][bus]["RF_inputs"]["port"]
                assert value == settings_config_dict["octaves"][port[0]]["RF_outputs"][port[1]]["gain"]
        if parameter == Parameter.TIME_OF_FLIGHT:
            if "time_of_flight" in config_keys:
                assert value == settings_config_dict["elements"][bus]["time_of_flight"]
        if parameter == Parameter.SMEARING:
            if "smearing" in config_keys:
                assert value == settings_config_dict["elements"][bus]["smearing"]

        if parameter == Parameter.THRESHOLD_ROTATION:
            element = next((element for element in qmm.settings.elements if element["identifier"] == bus))
            assert value == element.get("threshold_rotation", None)
        if parameter == Parameter.THRESHOLD:
            element = next((element for element in qmm.settings.elements if element["identifier"] == bus))
            assert value == element.get("threshold", None)

        if parameter == Parameter.DC_OFFSET:
            assert value == settings_config_dict["controllers"]["con1"]["analog_outputs"][5]["offset"]
        if parameter == Parameter.OFFSET_I:
            assert value == settings_config_dict["controllers"]["con1"]["analog_outputs"][3]["offset"]
        if parameter == Parameter.OFFSET_Q:
            assert value == settings_config_dict["controllers"]["con1"]["analog_outputs"][2]["offset"]
        if parameter == Parameter.OFFSET_OUT1:
            assert value == settings_config_dict["controllers"]["con1"]["analog_inputs"][1]["offset"]
        if parameter == Parameter.OFFSET_OUT2:
            assert value == settings_config_dict["controllers"]["con1"]["analog_inputs"][2]["offset"]

        # Assert that the settings are in synch:
        assert qmm._config_created is False and "_config" not in dir(qmm)

    @pytest.mark.parametrize(
        "bus, parameter, qmm_name",
        [
            ("flux_q0", Parameter.DC_OFFSET, "qmm_with_opx1000"),
            ("readout_q0_rf", Parameter.OFFSET_I, "qmm_with_opx1000"),
            ("drive_q0_rf", Parameter.OFFSET_Q, "qmm_with_opx1000"),
            ("readout_q0_rf", Parameter.OFFSET_OUT1, "qmm_with_opx1000"),
            ("readout_q0_rf", Parameter.OFFSET_OUT2, "qmm_with_opx1000"),
        ],
    )
    @patch("qililab.instruments.quantum_machines.quantum_machines_cluster.QuantumMachinesManager")
    @patch("qililab.instruments.quantum_machines.quantum_machines_cluster.QuantumMachine")
    def test_get_parameter_method_opx1000(
        self,
        mock_qmm,
        mock_qm,
        bus: str,
        parameter: Parameter,
        qmm_name: QuantumMachinesCluster,
        request,
    ):
        """Test the setup method with float value"""
        qmm = request.getfixturevalue(qmm_name)
        value = qmm.get_parameter(parameter, channel_id=bus)

        settings_config_dict = qmm.settings.to_qua_config()

        if parameter == Parameter.DC_OFFSET:
            assert value == settings_config_dict["controllers"]["con1"]["fems"][1]["analog_outputs"][5]["offset"]
        if parameter == Parameter.OFFSET_I:
            assert value == settings_config_dict["controllers"]["con1"]["fems"][1]["analog_outputs"][3]["offset"]
        if parameter == Parameter.OFFSET_Q:
            assert value == settings_config_dict["controllers"]["con1"]["fems"][1]["analog_outputs"][2]["offset"]
        if parameter == Parameter.OFFSET_OUT1:
            assert value == settings_config_dict["controllers"]["con1"]["fems"][1]["analog_inputs"][1]["offset"]
        if parameter == Parameter.OFFSET_OUT2:
            assert value == settings_config_dict["controllers"]["con1"]["fems"][1]["analog_inputs"][2]["offset"]

        # Assert that the settings are in synch:
        assert qmm._config_created is False and "_config" not in dir(qmm)

    @pytest.mark.parametrize("parameter", [(Parameter.MAX_CURRENT), (Parameter.OUT0_ATT)])
    @patch("qililab.instruments.quantum_machines.quantum_machines_cluster.QuantumMachinesManager")
    @patch("qililab.instruments.quantum_machines.quantum_machines_cluster.QuantumMachine")
    def test_get_parameter_method_raises_exception_when_parameter_not_found(
        self, mock_qmm, mock_qm, parameter: Parameter, qmm: QuantumMachinesCluster
    ):
        """Test the get_parameter method raises exception when parameter is wrong."""
        qmm.initial_setup()
        qmm.turn_on()
        with pytest.raises(ParameterNotFound):
            qmm.get_parameter(parameter, "drive_q0")

    @pytest.mark.parametrize("bus, parameter", [("drive_q0", Parameter.LO_FREQUENCY)])
    @patch("qililab.instruments.quantum_machines.quantum_machines_cluster.QuantumMachinesManager")
    @patch("qililab.instruments.quantum_machines.quantum_machines_cluster.QuantumMachine")
    def test_get_parameter_after_initial_setup(
        self, mock_qmm, mock_qm, bus: str, parameter: Parameter, qmm: QuantumMachinesCluster
    ):
        """Test the get_parameter method after and initial_setup."""

        qmm.initial_setup()
        value = qmm.get_parameter(parameter, channel_id=bus)
        config_keys = qmm._config["elements"][bus]

        if parameter == Parameter.LO_FREQUENCY:
            if "mixInputs" in config_keys:
                assert value == qmm._config["elements"][bus]["mixInputs"]["lo_frequency"]
            if "RF_inputs" in config_keys:
                port = qmm._config["elements"][bus]["RF_inputs"]["port"]
                assert value == qmm._config["octaves"][port[0]]["RF_outputs"][port[1]]["LO_frequency"]

        # Assert that the settings are still in synch:
        assert qmm._config == qmm.settings.to_qua_config()

    @pytest.mark.parametrize("bus, parameter", [("drive_q0", Parameter.LO_FREQUENCY)])
    @patch("qililab.instruments.quantum_machines.quantum_machines_cluster.QuantumMachinesManager")
    @patch("qililab.instruments.quantum_machines.quantum_machines_cluster.QuantumMachine")
    def test_get_parameter_doesnt_create_a_config(
        self, mock_qmm, mock_qm, bus: str, parameter: Parameter, qmm: QuantumMachinesCluster
    ):
        """Test the get_parameter method doesn't create a `_config`."""
        assert qmm._config_created is False
        qmm.get_parameter(parameter, channel_id=bus)
        assert qmm._config_created is False



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Yokogawa GS200 Instrument"""

from dataclasses import dataclass

from qililab.instruments import InstrumentFactory, ParameterNotFound, check_device_initialized, log_set_parameter
from qililab.instruments.current_source import CurrentSource
from qililab.instruments.voltage_source import VoltageSource
from qililab.typings import ChannelID, InstrumentName, Parameter, ParameterValue
from qililab.typings import YokogawaGS200 as YokogawaGS200Driver
from qililab.typings.enums import SourceMode


@InstrumentFactory.register
class GS200(CurrentSource, VoltageSource):
    """Yokogawa GS200 low voltage/current DC source.

    Args:
        name (InstrumentName): name of the instrument
        device (YokogawaGS200Driver): Instance of the qcodes GS200 class.
        settings (YokogawaGS200Settings): Settings of the instrument.
    """

    name = InstrumentName.YOKOGAWA_GS200

    @dataclass
    class YokogawaGS200Settings(CurrentSource.CurrentSourceSettings, VoltageSource.VoltageSourceSettings):
        """Contains the settings of a specific signal generator."""

        source_mode: SourceMode

    settings: YokogawaGS200Settings
    device: YokogawaGS200Driver

    @property
    def source_mode(self) -> SourceMode:
        """Yokogawa GS200 `source_mode` property. It determines if the instrument will act as Current Source or Voltage Source.

        Returns:
            SourceMode: The source mode of the instrument.
        """
        return self.settings.source_mode

    @source_mode.setter
    def source_mode(self, value: SourceMode):
        """Set the Yokogawa GS200 source mode."""
        self.settings.source_mode = value
        if self.is_device_active():
            qcodes_str = self.__source_mode_to_qcodes_str(value)
            self.device.source_mode(qcodes_str)

    @property
    def ramping_enabled(self):
        """Yokogawa GS200 'ramping_enabled' property. If set to True, changes in current will transition linearly according to the `ramping_rate` value.

        Returns:
            bool: True, if ramping is enabled.
        """
        return self.settings.ramping_enabled[0]

    @ramping_enabled.setter
    def ramping_enabled(self, value: bool):
        """Set the Yokogawa `ramping_enabled` property."""
        self.settings.ramping_enabled[0] = value

    @property
    def ramping_rate(self):
        """Yokogawa GS200 'ramping_rate' property. If `ramping_enabled` is set to True, changes in current will transition linearly according to this value.

        Returns:
            float: The ramping rate in mA/ms or mV/ms.
        """
        return self.settings.ramp_rate[0]

    @ramping_rate.setter
    def ramping_rate(self, value: float):
        """Set the Yokogawa GS200 `ramping_rate` property."""
        self.settings.ramp_rate[0] = value

    @property
    def current(self) -> float:
        """Yokogawa GS200 `current` property.

        Returns:
            float: The current of the instrument in A.
        """
        return self.settings.current[0]

    @current.setter
    def current(self, value: float):
        """Set Yokogawa GS200 `current` property. If `ramping_enabled` is set to True, the current will transition linearly according to `ramping_rate`. Else, it will be set instantly."""
        self.settings.current[0] = value
        if self.is_device_active():
            if self.ramping_enabled:
                self.device.ramp_current(value, self.ramping_rate * 0.001, 0.001)
            else:
                self.device.current(value)

    @property
    def span(self) -> str:
        """Yokogawa GS200 `span` property.

        Returns:
            str: The span of the instrument.
        """
        return self.settings.span[0]

    @span.setter
    def span(self, value: str):
        """Set Yokogawa GS200 `span` property.

        Available values:
            - 10mV
            - 100mV
            - 1V
            - 10V
            - 30V
            - 1mA
            - 10mA
            - 100mA
            - 200mA
        """
        self.settings.span[0] = value
        if self.is_device_active():
            if self.source_mode == SourceMode.CURRENT:
                qcodes_int = self.__current_span_to_qcodes_float(value)
                self.device.current_range(qcodes_int)
            else:
                qcodes_int = self.__voltage_span_to_qcodes_float(value)
                self.device.voltage_range(qcodes_int)

    @property
    def voltage(self) -> float:
        """Yokogawa GS200 `voltage` property.

        Returns:
            float: The voltage of the instrument in V.
        """
        return self.settings.voltage[0]

    @voltage.setter
    def voltage(self, value: float):
        """Set Yokogawa GS200 `voltage` property. If `ramping_enabled` is set to True, the voltage will transition linearly according to `ramping_rate`. Else, it will be set instantly."""
        self.settings.voltage[0] = value
        if self.is_device_active():
            if self.ramping_enabled:
                self.device.ramp_voltage(value, self.ramping_rate * 0.001, 0.001)
            else:
                self.device.voltage(value)

    @log_set_parameter
    def set_parameter(self, parameter: Parameter, value: ParameterValue, channel_id: ChannelID | None = None):
        """Set instrument settings parameter to the corresponding value

        Args:
            parameter (Parameter): settings parameter to be updated
            value (float | str | bool): new value
            channel_id (int | None): channel identifier of the parameter to update
        """
        if parameter == Parameter.CURRENT:
            self.current = float(value)
            return

        if parameter == Parameter.VOLTAGE:
            self.voltage = float(value)
            return

        if parameter == Parameter.RAMPING_ENABLED:
            self.ramping_enabled = bool(value)
            return

        if parameter == Parameter.RAMPING_RATE:
            self.ramping_rate = float(value)
            return

        if parameter == Parameter.SOURCE_MODE:
            self.source_mode = SourceMode(value)
            return

        if parameter == Parameter.SPAN:
            self.span = str(value)
            return

        raise ParameterNotFound(self, parameter)

    def get_parameter(self, parameter: Parameter, channel_id: ChannelID | None = None):
        if parameter == Parameter.CURRENT:
            return float(self.current)

        if parameter == Parameter.VOLTAGE:
            return float(self.voltage)

        if parameter == Parameter.RAMPING_ENABLED:
            return bool(self.ramping_enabled)

        if parameter == Parameter.RAMPING_RATE:
            return float(self.ramping_rate)

        if parameter == Parameter.SOURCE_MODE:
            return SourceMode(self.source_mode)

        if parameter == Parameter.SPAN:
            return str(self.span)

        raise ParameterNotFound(self, parameter)

    @check_device_initialized
    def initial_setup(self):
        """Performs an initial setup."""
        self.device.off()
        self.source_mode = self.settings.source_mode
        self.span = self.settings.span[0]
        self.ramping_rate = self.settings.ramp_rate[0]
        self.ramping_enabled = self.settings.ramping_enabled[0]
        if self.source_mode == SourceMode.CURRENT:
            self.current = self.settings.current[0]
        else:
            self.voltage = self.settings.voltage[0]

    @check_device_initialized
    def turn_on(self):
        """Start outputting current."""
        self.device.on()

    @check_device_initialized
    def turn_off(self):
        """Stop outputing current."""
        self.device.off()

    @check_device_initialized
    def reset(self):
        """Stop outputing current."""

    def __source_mode_to_qcodes_str(self, source_mode: SourceMode) -> str:
        mapping = {SourceMode.CURRENT: "CURR", SourceMode.VOLTAGE: "VOLT"}
        return mapping[source_mode]

    def __voltage_span_to_qcodes_float(self, voltage_span: str) -> float:
        mapping = {"10mV": 10e-3, "100mV": 100e-3, "1V": 1e0, "10V": 10e0, "30V": 30e0}
        return mapping[voltage_span]

    def __current_span_to_qcodes_float(self, current_span: str) -> float:
        mapping = {"1mA": 1e-3, "10mA": 10e-3, "100mA": 100e-3, "200mA": 200e-3}
        return mapping[current_span]



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" Utils for dictionaries manipulation """
import copy


def merge_dictionaries(origin: dict, new: dict) -> dict:
    """Recursively merge `new` into a copy of `origin` and returns the new dictionary."""
    merged = copy.deepcopy(origin)

    for key, value in new.items():
        if key in merged and isinstance(merged[key], dict) and isinstance(value, dict):
            # Recursively merge dictionaries
            merged[key] = merge_dictionaries(merged[key], value)
        else:
            merged[key] = value

    return merged



"""Module containing utilities for the tests."""

import copy
from unittest.mock import MagicMock, patch

import numpy as np
from qcodes.instrument_drivers.tektronix.Keithley_2600_channels import KeithleyChannel
from qpysequence import Sequence as QPySequence
from qpysequence.program import Program as QPyProgram

import qililab as ql
from qililab.platform import Platform


def mock_instruments(mock_rs: MagicMock, mock_pulsar: MagicMock, mock_keithley: MagicMock):
    """Mock dynamically created attributes."""
    mock_rs_instance = mock_rs.return_value
    mock_rs_instance.mock_add_spec(["power", "frequency", "ref_osc_source"])
    mock_pulsar_instance = mock_pulsar.return_value
    mock_pulsar_instance.get_acquisitions.side_effect = lambda sequencer: copy.deepcopy(
        {
            "acq_q0_0": {
                "index": 0,
                "acquisition": {
                    "scope": {
                        "path0": {"data": [1, 1, 1, 1, 1, 1, 1, 1], "out-of-range": False, "avg_cnt": 1},
                        "path1": {"data": [0, 0, 0, 0, 0, 0, 0, 0], "out-of-range": False, "avg_cnt": 1},
                    },
                    "bins": {
                        "integration": {"path0": [-0.08875841551660968], "path1": [-0.4252879595139228]},
                        "threshold": [0],
                        "avg_cnt": [1],
                    },
                },
            }
        }
    )
    mock_pulsar_instance.mock_add_spec(
        [
            "reference_source",
            "sequencer0",
            "out0_offset",
            "out1_offset",
            "out2_offset",
            "out3_offset",
            "scope_acq_avg_mode_en_path0",
            "scope_acq_avg_mode_en_path1",
            "scope_acq_trigger_mode_path0",
            "scope_acq_trigger_mode_path1",
            "sequencers",
            "scope_acq_sequencer_select",
            "get_acquisitions",
            "disconnect_outputs",
            "disconnect_inputs",
        ]
    )
    mock_pulsar_instance.sequencer0.mock_add_spec(
        [
            "sync_en",
            "gain_awg_path0",
            "gain_awg_path1",
            "sequence",
            "mod_en_awg",
            "nco_freq",
            "scope_acq_sequencer_select",
            "channel_map_path0_out0_en",
            "channel_map_path1_out1_en",
            "demod_en_acq",
            "integration_length_acq",
            "mixer_corr_phase_offset_degree",
            "mixer_corr_gain_ratio",
        ]
    )
    mock_keithley_instance = mock_keithley.return_value
    mock_keithley_instance.smua = MagicMock(KeithleyChannel)
    mock_keithley_instance.smua.mock_add_spec(["limiti", "limitv", "doFastSweep"])


def name_generator(base: str):
    """Unique name generator
    Args:
        base (str): common name for all the elements.
    Yields:
        str: Unique name in the format base_id.
    """
    next_id = 0
    while True:
        yield f"{base}_{next_id!s}"
        next_id += 1


def compare_pair_of_arrays(
    pair_a: tuple[list[float], list[float]],
    pair_b: tuple[list[float], list[float]],
    tolerance: float,
) -> bool:
    """Compares two pairs of arrays of the same length up to a given tolerance.

    Args:
        pair_a (tuple[list[float], list[float]]): First pair of arrays.
        pair_b (tuple[list[float], list[float]]): Second pair of arrays.
        tolerance (float): Absolute amount up to which the arrays can differ to be considered equal.

    Returns:
        bool: True if the arrays are equal up to the given tolerance, False otherwise.
    """
    path0_ok = all(np.isclose(pair_a[0], pair_b[0], atol=tolerance))
    path1_ok = all(np.isclose(pair_a[1], pair_b[1], atol=tolerance))
    return path0_ok and path1_ok


def complete_array(array: list[float], filler: float, final_length: int) -> list[float]:
    """Fills a given array with the given float as a filler up to the final_length specified.

    Args:
        array (list[float]): Original array.
        filler (float): Number to use as a filler.
        final_length (int): Final length of the array.

    Returns:
        list[float]: List of length `final_length` where the first `len(array)` elements are those of the original
            array, and the remaining elements are are repetitions of `filler`.
    """
    return array + [filler] * (final_length - len(array))


dummy_qrm_name_generator = name_generator("dummy_qrm")
dummy_qcm_name_generator = name_generator("dummy_qcm")


def build_platform(runcard: dict) -> Platform:
    """Return PlatformBuilderDB instance with loaded platform."""
    runcard = copy.deepcopy(runcard)
    with patch("ruamel.yaml.YAML.load", return_value=runcard) as mock_load:
        with patch("qililab.data_management.open") as mock_open:
            pl = ql.build_platform(runcard="_")
            mock_load.assert_called()
            mock_open.assert_called()
    return pl


def is_q1asm_equal(a: str | QPySequence | QPyProgram, b: str | QPySequence | QPyProgram) -> bool:
    if isinstance(a, QPySequence):
        a = repr(a._program)
    if isinstance(a, QPyProgram):
        a = repr(a)

    if isinstance(b, QPySequence):
        b = repr(b._program)
    if isinstance(b, QPyProgram):
        b = repr(b)
    return "".join(a.strip().split()) == "".join(b.strip().split())



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys

# For Python 3.11+, we use ExceptionGroup; otherwise, we define a custom exception class
if sys.version_info >= (3, 11):  # pragma: no cover
    # Available natively in Python 3.11+
    from builtins import ExceptionGroup
else:
    # Define a custom exception class for earlier versions
    class ExceptionGroup(Exception):
        """Custom implementation of ExceptionGroup."""

        def __init__(self, message, exceptions):
            self.exceptions = exceptions
            super().__init__(f"{message}: {', '.join(str(e) for e in exceptions)}")
